#!/usr/bin/env python
# SPDX-License-Identifier: ISC
# -*- coding: utf-8 eval: (blacken-mode 1) -*-
#
# September 17 2024, Christian Hopps <chopps@labn.net>
#
# Copyright (c) 2024, LabN Consulting, L.L.C.
#

import json
import pytest
from lib.topogen import Topogen

pytestmark = [pytest.mark.staticd, pytest.mark.mgmtd]


@pytest.fixture(scope="module")
def tgen(request):
    "Setup/Teardown the environment and provide tgen argument to tests"

    topodef = {"s1": ("r1",)}

    tgen = Topogen(topodef, request.module.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    for rname, router in router_list.items():
        router.load_frr_config("frr-yanglib.conf")

    tgen.start_router()
    yield tgen
    tgen.stop_topology()


def test_yang_lib(tgen):
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"].net
    output = r1.cmd_nostatus(
        "vtysh -c 'show mgmt get-data /ietf-yang-library:yang-library'"
    )
    ret = json.loads(output)
    loaded_modules = ret['ietf-yang-library:yang-library']['module-set'][0]['module']
    assert len(loaded_modules) > 10, "Modules missing from yang-library"



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright (c) 2021 by
# Donatas Abraitis <donatas.abraitis@gmail.com>
#

"""
Test if BGP Long-lived Graceful Restart capability works:
    Check if we can see 172.16.1.1/32 after initial converge in R3.
    Check if we can see 172.16.1.1/32 as best selected due to higher weigth in R2.
    Kill bgpd in R1.
    Check if we can see 172.16.1.1/32 as stale in R2.
    Check if we can see 172.16.1.1/32 depreferenced due to LLGR_STALE in R2.
    Check if we can see 172.16.1.1/32 after R1 was killed in R3.
"""

import os
import sys
import json
import pytest
import functools

pytestmark = [pytest.mark.bgpd]

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen

from lib.common_config import (
    kill_router_daemons,
    step,
)


def build_topo(tgen):
    for routern in range(0, 6):
        tgen.add_router("r{}".format(routern))

    switch = tgen.add_switch("s0")
    switch.add_link(tgen.gears["r0"])
    switch.add_link(tgen.gears["r2"])

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])

    # Dynamic neighbor
    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r4"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_llgr():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]
    r3 = tgen.gears["r3"]

    def _bgp_converge(router):
        output = json.loads(router.vtysh_cmd("show ip bgp json"))
        expected = {
            "routes": {
                "172.16.1.1/32": [{"nexthops": [{"ip": "192.168.2.2", "used": True}]}],
                "172.16.1.2/32": [{"nexthops": [{"ip": "192.168.2.2", "used": True}]}],
            }
        }
        return topotest.json_cmp(output, expected)

    step("Check if we can see 172.16.1.1/32 after initial converge in R3")
    test_func = functools.partial(_bgp_converge, r3)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Cannot see 172.16.1.1/32 in r3"

    def _bgp_weight_prefered_route(router):
        output = json.loads(router.vtysh_cmd("show ip bgp 172.16.1.1/32 json"))
        expected = {
            "paths": [
                {
                    "bestpath": {"selectionReason": "Weight"},
                    "nexthops": [
                        {
                            "ip": "192.168.1.1",
                        }
                    ],
                }
            ]
        }
        return topotest.json_cmp(output, expected)

    step(
        "Check if we can see 172.16.1.1/32 as best selected due to higher weigth in R2"
    )
    test_func = functools.partial(_bgp_weight_prefered_route, r2)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert (
        result is None
    ), "Prefix 172.16.1.1/32 is not selected as bests path due to weight"

    step("Kill bgpd in R1")
    kill_router_daemons(tgen, "r1", ["bgpd"])

    def _bgp_stale_route(router, prefix):
        output = json.loads(router.vtysh_cmd("show ip bgp {} json".format(prefix)))
        expected = {"paths": [{"community": {"string": "llgr-stale"}, "stale": True}]}
        return topotest.json_cmp(output, expected)

    step("Check if we can see 172.16.1.1/32 as stale in R2")
    test_func = functools.partial(_bgp_stale_route, r2, "172.16.1.1/32")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Prefix 172.16.1.1/32 is not stale"

    def _bgp_llgr_depreference_route(router):
        output = json.loads(router.vtysh_cmd("show ip bgp 172.16.1.1/32 json"))
        expected = {
            "paths": [
                {
                    "bestpath": {"selectionReason": "First path received"},
                    "nexthops": [
                        {
                            "ip": "192.168.0.1",
                        }
                    ],
                }
            ]
        }
        return topotest.json_cmp(output, expected)

    step("Check if we can see 172.16.1.1/32 depreferenced due to LLGR_STALE in R2")
    test_func = functools.partial(_bgp_llgr_depreference_route, r2)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Prefix 172.16.1.1/32 is not depreferenced due to LLGR_STALE"

    step("Check if we can see 172.16.1.1/32 after R1 was killed in R3")
    test_func = functools.partial(_bgp_converge, r3)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Cannot see 172.16.1.1/32 in r3"

    step("Kill bgpd in R4 (dynamic peer)")
    kill_router_daemons(tgen, "r4", ["bgpd"])

    step("Check if we can see 172.16.1.2/32 after R4 (dynamic peer) was killed")
    test_func = functools.partial(_bgp_stale_route, r2, "172.16.1.2/32")
    _, result = topotest.run_and_expect(test_func, None, count=120, wait=0.5)
    assert result is None, "Cannot see 172.16.1.2/32 in r2"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2020 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

"""
Following tests are covered to test bgp community functionality:
1. Verify that BGP well known communities work fine for
   eBGP and iBGP peers.
   Well known communities tested: no-export, local-AS

"""

import os
import sys
import time
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen

# Import topoJson from lib, to create topology and initial configuration
from lib.common_config import (
    start_topology,
    write_test_header,
    write_test_footer,
    reset_config_on_routers,
    verify_rib,
    create_static_routes,
    check_address_types,
    step,
    create_route_maps,
    create_route_maps,
    required_linux_kernel_version,
)

from lib.topolog import logger
from lib.bgp import (
    verify_bgp_convergence,
    create_router_bgp,
    verify_bgp_rib,
    verify_bgp_community,
)
from lib.topojson import build_config_from_json

pytestmark = [pytest.mark.bgpd, pytest.mark.staticd]


# Global variables
BGP_CONVERGENCE = False
ADDR_TYPES = check_address_types()
NETWORK = {
    "ipv4": ["192.0.2.1/32", "192.0.2.2/32"],
    "ipv6": ["2001:DB8::1:1/128", "2001:DB8::1:2/128"],
}


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    # Required linux kernel version for this suite to run.
    result = required_linux_kernel_version("4.14")
    if result is not True:
        pytest.skip("Kernel requirements are not met, kernel version should be >= 4.14")

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/bgp_communities_topo2.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # Checking BGP convergence
    global BGP_CONVERGENCE
    global ADDR_TYPES

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Api call verify whether BGP is converged
    BGP_CONVERGENCE = verify_bgp_convergence(tgen, topo)
    assert BGP_CONVERGENCE is True, "setup_module :Failed \n Error:" " {}".format(
        BGP_CONVERGENCE
    )

    logger.info("Running setup_module() done")


def teardown_module(mod):
    """
    Teardown the pytest environment

    * `mod`: module name
    """

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


#####################################################
#
#   Tests starting
#
#####################################################


def test_bgp_no_export_local_as_communities_p0(request):
    """
    Verify that BGP well known communities work fine for
    eBGP and iBGP peers.
    Well known communities tested: no-export, local-AS
    """

    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Initial config: Configure BGP neighborship between R1 and R3.")
    reset_config_on_routers(tgen)

    step("Configure static routes on R1 with next-hop as null0")
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r1": {
                "static_routes": [{"network": NETWORK[addr_type], "next_hop": "null0"}]
            }
        }
        result = create_static_routes(tgen, input_dict_4)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    for comm_type in ["no-export", "local-AS"]:
        step("Create a route-map on R1 to set community as {}".format(comm_type))

        seq_id = 10
        input_rmap = {
            "r1": {
                "route_maps": {
                    "rmap_wkc": [
                        {
                            "action": "permit",
                            "seq_id": seq_id,
                            "set": {"community": {"num": "{}".format(comm_type)}},
                        }
                    ]
                }
            }
        }
        result = create_route_maps(tgen, input_rmap)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        step("Apply route-map while redistributing static routes into BGP")
        input_dict_2 = {
            "r1": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "redistribute": [
                                    {
                                        "redist_type": "static",
                                        "attribute": {"route-map": "rmap_wkc"},
                                    }
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "redistribute": [
                                    {
                                        "redist_type": "static",
                                        "attribute": {"route-map": "rmap_wkc"},
                                    }
                                ]
                            }
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_2)

        step("Verify that BGP prefixes on R1 have community: {}".format(comm_type))
        input_dict_4 = {"community": "{}".format(comm_type)}
        for addr_type in ADDR_TYPES:
            result = verify_bgp_community(
                tgen, addr_type, "r1", NETWORK[addr_type], input_dict_4
            )
            assert result is True, "Test case {} : Should fail \n Error: {}".format(
                tc_name, result
            )

        for addr_type in ADDR_TYPES:
            input_dict_4 = {
                "r1": {
                    "static_routes": [
                        {
                            "network": NETWORK[addr_type],
                            "next_hop": topo["routers"]["r2"]["links"]["r1"][
                                addr_type
                            ].split("/")[0],
                        }
                    ]
                }
            }
            result = verify_bgp_rib(
                tgen,
                addr_type,
                "r2",
                input_dict_4,
                next_hop=topo["routers"]["r1"]["links"]["r2"][addr_type].split("/")[0],
            )
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )

        step(
            "Verify that these prefixes, originated on R1, are not"
            "received on R3 but received on R2"
        )
        result = verify_rib(
            tgen,
            addr_type,
            "r3",
            input_dict_4,
            next_hop=topo["routers"]["r1"]["links"]["r3"][addr_type].split("/")[0],
            expected=False,
        )
        assert result is not True, (
            "Testcase {} : Failed \n "
            "Expected: Routes are still present in rib of r3 \n "
            "Found: {}".format(tc_name, result)
        )

        step("Remove route-map from redistribute static on R1")
        input_dict_2 = {
            "r1": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static", "delete": True}
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static", "delete": True}
                                ]
                            }
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_2)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        step("Configure redistribute static")
        input_dict_2 = {
            "r1": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {"redistribute": [{"redist_type": "static"}]}
                        },
                        "ipv6": {
                            "unicast": {"redistribute": [{"redist_type": "static"}]}
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_2)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        step(
            "Verify that these prefixes, originated on R1, are now"
            "received on both routers R2 and R3"
        )
        for addr_type in ADDR_TYPES:
            input_dict_4 = {
                "r1": {
                    "static_routes": [
                        {
                            "network": NETWORK[addr_type],
                            "next_hop": topo["routers"]["r2"]["links"]["r1"][
                                addr_type
                            ].split("/")[0],
                        }
                    ]
                }
            }
            result = verify_bgp_rib(
                tgen,
                addr_type,
                "r2",
                input_dict_4,
                next_hop=topo["routers"]["r1"]["links"]["r2"][addr_type].split("/")[0],
            )
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )

            result = verify_bgp_rib(
                tgen,
                addr_type,
                "r3",
                input_dict_4,
                next_hop=topo["routers"]["r1"]["links"]["r3"][addr_type].split("/")[0],
            )
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright (c) 2024 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, get_topogen

pytestmark = [pytest.mark.bgpd]


def setup_module(mod):
    topodef = {"s1": ("r1", "r2")}
    tgen = Topogen(topodef, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_frr_config(os.path.join(CWD, "{}/frr.conf".format(rname)))

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_received_routes_with_soft_inbound():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"]

    def _bgp_converge():
        output = json.loads(r1.vtysh_cmd("show bgp ipv4 unicast json"))
        expected = {
            "routes": {
                "10.0.0.2/32": [
                    {
                        "valid": True,
                        "path": "65000 65000 65000 65002",
                        "nexthops": [
                            {
                                "ip": "192.168.1.2",
                            }
                        ],
                    }
                ]
            }
        }

        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_converge,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "Can't converge"

    def _bgp_check_receveived_routes():
        output = json.loads(
            r1.vtysh_cmd(
                "show bgp ipv4 unicast neighbors 192.168.1.2 received-routes json"
            )
        )
        expected = {
            "receivedRoutes": {
                "10.0.0.2/32": {
                    "valid": True,
                    "path": "65002",
                    "nextHop": "192.168.1.2",
                }
            }
        }

        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_check_receveived_routes,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "Can't converge"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2023 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

"""
Check if IPv6 Link-Local BGP peering works fine.
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen

pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    for routern in range(1, 4):
        tgen.add_router("r{}".format(routern))

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r3"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_ipv6_link_local_peering():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"]
    r3 = tgen.gears["r3"]

    def _bgp_converge():
        output = json.loads(r1.vtysh_cmd("show bgp summary json"))
        expected = {
            "ipv4Unicast": {
                "peers": {
                    "fe80:1::2": {
                        "state": "Established",
                    }
                }
            }
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_converge)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see BGP convergence on R2"

    def _bgp_router_id_missing():
        output = json.loads(r3.vtysh_cmd("show bgp summary failed json"))
        expected = {
            "ipv4Unicast": {
                "routerId": "0.0.0.0",
                "as": 65003,
                "peers": {
                    "fe80:1::1": {
                        "connectionsEstablished": 0,
                        "connectionsDropped": 0,
                        "peerUptime": "never",
                        "lastResetDueTo": "Router ID is missing",
                    }
                },
            }
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_router_id_missing)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "r3 should stay down due to missing router ID"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_bgp_vpnv4_asbr.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2023 by 6WIND
#

"""
 test_bgp_vpnv4_asbr.py: Test the FRR BGP daemon with rfc4364 option 10b
 r1, r2, and r100 are in an iBGP AS, while r2, r3 do an eBGP peering
 h1 is a host behind r1 VRF1, and {h2,h3} are hosts behind r3 VRF1
 The test demonstrates the connectivity across the network between h1 and h3.


 +----------+          +----+--------+              +--------+              +--------+-----+
 |          |172.31.0.0|vrf | r1     |192.168.0.0/24| r2     |192.168.1.0/24|r3      | vrf |
 |   h1     +----------+    |       1+------+-------+        +------+-------+3       |     +--- 172.31.3.0/24
 |   10     |          |VRF1|AS65500 |      |       | AS65500|      |       |AS65501 |VRF1 |
 +----------+          +-------------+      |       +--------+      |       +--------+--+-++
                           192.0.2.1        |       192.0.2.2       |                172| |
                                       +----------+            +----+--------+        31| |
                                       |rr100     |            |rs200/AS65502|         1| |
                                       +----------+            +-------------+         0| |
                                        192.0.2.100   +--------+                     /24| |
                                                      |        |        +----------+----+ |
                                                      |h3      |        |          |      |
                                                      |10      |        |  h2      |      |
                                                      +---+----+        |  10      |      |
                                                          |             +----------+      |
                                                          |172.31.2.0/24                  |
                                                          +--------------------------------+
"""

import os
import sys
import json
from functools import partial
import pytest
import functools

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.bgpcheck import (
    check_show_bgp_vpn_prefix_found,
    check_show_bgp_vpn_prefix_not_found,
)
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger
from lib.checkping import check_ping


# Required to instantiate the topology builder class.


pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    "Build function"

    # Allocate 8 devices
    tgen.add_router("r1")
    tgen.add_router("r2")
    tgen.add_router("r3")
    tgen.add_router("h1")
    tgen.add_router("h2")
    tgen.add_router("h3")
    tgen.add_router("rr100")
    tgen.add_router("rs200")

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["rr100"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["h1"])

    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])
    switch.add_link(tgen.gears["rs200"])

    switch = tgen.add_switch("s4")
    switch.add_link(tgen.gears["r3"])
    switch.add_link(tgen.gears["h2"])

    switch = tgen.add_switch("s5")
    switch.add_link(tgen.gears["r3"])
    switch.add_link(tgen.gears["h3"])

    switch = tgen.add_switch("s6")
    switch.add_link(tgen.gears["r3"])


def _populate_iface():
    tgen = get_topogen()
    cmds_list = [
        "ip link add vrf1 type vrf table 10",
        "echo 100000 > /proc/sys/net/mpls/platform_labels",
        "ip link set dev vrf1 up",
        "ip link set dev {0}-eth1 master vrf1",
        "echo 1 > /proc/sys/net/mpls/conf/{0}-eth0/input",
    ]

    for rname in ("r1", "r3"):
        for cmd in cmds_list:
            input = cmd.format(rname)
            logger.info("input: " + cmd)
            output = tgen.net[rname].cmd(cmd.format(rname))
            logger.info("output: " + output)

    cmds_list = [
        "ip link set dev {0}-eth2 master vrf1",
        "ip link set dev {0}-eth3 master vrf1",
    ]
    for cmd in cmds_list:
        input = cmd.format("r3")
        logger.info("input: " + input)
        output = tgen.net["r3"].cmd(input)
        logger.info("output: " + output)


def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    _populate_iface()

    for rname, router in router_list.items():
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        if rname in ("r1", "r2", "r3", "rr100", "rs200"):
            router.load_config(
                TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
            )

    # Initialize all routers.
    tgen.start_router()


def teardown_module(_mod):
    "Teardown the pytest environment"
    tgen = get_topogen()

    tgen.stop_topology()


def bgp_vpnv4_prefix_check(router, rd, prefix, label, nexthop):
    """
    Dump and check 'show bgp ipv4 vpn <prefix> json' output. An assert is triggered in case test fails
    * 'router': the router to check
    * 'rd': The route distinguisher expected
    * 'prefix': The prefix expected
    * 'label': The label expected associated with the ('rd','prefix') tuple
    * 'nexthop': The nexthop expected associated with the ('rd','prefix') tuple
    """

    def _check(router, prefix, rd, label, nexthop):
        dump = router.vtysh_cmd("show bgp ipv4 vpn {} json".format(prefix), isjson=True)
        if not dump:
            return "{0}, {1}, route distinguisher {2} not present".format(
                router.name, prefix, rd
            )
        for dumped_rd, pathes in dump.items():
            if dumped_rd != rd:
                continue
            for path in pathes["paths"]:
                if "remoteLabel" not in path.keys():
                    return "{0}, {1}, rd {2}, remoteLabel not present".format(
                        router.name, prefix, rd
                    )
                if str(path["remoteLabel"]) != label:
                    continue

                if "nexthops" not in path.keys():
                    return "{0}, {1}, rd {2}, no nexthops present".format(
                        router.name, prefix, rd
                    )

                for nh in path["nexthops"]:
                    if "ip" not in nh.keys():
                        return "{0}, {1}, rd {2}, no ipv4 nexthop available".format(
                            router.name, prefix, rd
                        )
                    if nh["ip"] != nexthop:
                        continue
                    return None
        return "{0}, {1}, rd {2}, remoteLabel {3}, nexthop {4} not found".format(
            router.name, prefix, rd, label, nexthop
        )

    func = functools.partial(_check, router, prefix, rd, label, nexthop)
    _, result = topotest.run_and_expect(func, None, count=20, wait=0.5)
    assert_msg = "{}, show bgp ipv4 vpn {}, rd {}, label {} nexthop {}".format(
        router.name, prefix, rd, label, nexthop
    )
    assert result is None, assert_msg + " not found"
    logger.info(assert_msg + " found")


def mpls_table_get_entry(router, out_label, out_nexthop):
    """
    Get the in_label from tuple (out_label, out_nexthop)
    * 'router': the router to check
    * 'out_label': The outgoing label expected
    * 'out_nexthop': The outgoing nexthop expected
    """
    dump = router.vtysh_cmd("show mpls table json", isjson=True)
    for in_label, label_info in dump.items():
        for nh in label_info["nexthops"]:
            if nh["type"] != "BGP" or "installed" not in nh.keys():
                continue
            if "nexthop" in nh.keys():
                if nh["nexthop"] != out_nexthop:
                    continue
            if "outLabelStack" in nh.keys():
                if out_label not in nh["outLabelStack"]:
                    continue
            return in_label
    return None


def mpls_table_check_entry(router, out_label, out_nexthop):
    """
    Dump and check 'show mpls table json' output. An assert is triggered in case test fails
    * 'router': the router to check
    * 'out_label': The outgoing label expected
    * 'out_nexthop': The outgoing nexthop expected
    """
    logger.info("Checking MPLS labels on {}".format(router.name))
    dump = router.vtysh_cmd("show mpls table json", isjson=True)
    for in_label, label_info in dump.items():
        for nh in label_info["nexthops"]:
            if nh["type"] != "BGP" or "installed" not in nh.keys():
                continue
            if "nexthop" in nh.keys():
                if nh["nexthop"] != out_nexthop:
                    continue
            if "outLabelStack" in nh.keys():
                if out_label not in nh["outLabelStack"]:
                    continue
            logger.info(
                "{}, show mpls table, entry in_label {} out_label {} out_nexthop {} found".format(
                    router.name, in_label, nh["outLabelStack"], nh["nexthop"]
                )
            )
            return None
    return "{}, show mpls table, entry matching in_label {} out_label {} out_nexthop {} not found".format(
        router.name, in_label, out_label, out_nexthop
    )


def check_show_mpls_table_entry_label_not_found(router, inlabel):
    output = json.loads(router.vtysh_cmd("show mpls table {} json".format(inlabel)))
    expected = {"inLabel": inlabel, "installed": True}
    ret = topotest.json_cmp(output, expected)
    if ret is None:
        return "not good"
    return None


def check_show_bgp_vpn_ok(router, vpnv4_entries):
    """
    Check on router that BGP l3vpn entries are present
    Check there is an MPLS entry bound to that BGP L3VPN entry
    Extract the Label value and check on the distributed router the BGP L3VPN entry
    If check fail, an assert is triggered.
    * 'router': the router to check BGP VPN RIB
    * 'vpnv4_entries': dictionary that contains the list of prefixes, and the distributed router to look after
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    vpnv4_nexthops = {"r1": "192.0.2.2", "r3": "192.168.1.2"}
    vpnv4_nht = {"192.0.2.1": "192.168.0.1", "192.168.1.3": "192.168.1.3"}
    label_ip_entries = {}

    def _return_remote_label_nh_rd(router, prefix):
        dump = router.vtysh_cmd("show bgp ipv4 vpn {} json".format(prefix), isjson=True)
        assert_msg = (
            "{}, prefix {} not available or label not found",
            router.name,
            prefix,
        )
        assert dump, assert_msg
        for rd, pathes in dump.items():
            for path in pathes["paths"]:
                if "remoteLabel" not in path.keys():
                    assert 0, assert_msg
                for nh in path["nexthops"]:
                    if "ip" in nh.keys():
                        return path["remoteLabel"], nh["ip"], rd
        assert 0, assert_msg

    def _check_nexthop_available(router, prefix):
        dump = router.vtysh_cmd("show bgp ipv4 vpn {} json".format(prefix), isjson=True)
        if not dump:
            return "{0}, {1}, route distinguisher not present".format(
                router.name, prefix
            )
        for _, pathes in dump.items():
            for path in pathes["paths"]:
                if "remoteLabel" not in path.keys():
                    return "{0}, {1}, remoteLabel not present".format(
                        router.name, prefix
                    )
                if "nexthops" not in path.keys():
                    return "{0}, {1}, no nexthop available".format(router.name, prefix)
                return None

    for prefix, rname_to_test in vpnv4_entries.items():
        func = functools.partial(_check_nexthop_available, router, prefix)
        _, result = topotest.run_and_expect(func, None, count=20, wait=0.5)
        assert result is None, "Failed to detect prefix {} on router {}".format(
            prefix, router.name
        )

    for prefix, rname_to_test in vpnv4_entries.items():
        l3vpn_label, l3vpn_nh, l3vpn_rd = _return_remote_label_nh_rd(router, prefix)
        logger.info(
            "{0}, {1}, label value is {2}, nh is {3}".format(
                router.name, prefix, l3vpn_label, l3vpn_nh
            )
        )
        test_func = functools.partial(
            mpls_table_check_entry, router, l3vpn_label, vpnv4_nht[l3vpn_nh]
        )
        success, result = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
        assert success, result

        in_label = mpls_table_get_entry(router, l3vpn_label, vpnv4_nht[l3vpn_nh])
        label_ip_entries[prefix] = in_label

        bgp_vpnv4_prefix_check(
            tgen.gears[rname_to_test],
            l3vpn_rd,
            prefix,
            in_label,
            vpnv4_nexthops[rname_to_test],
        )

        return label_ip_entries


def test_protocols_convergence():
    """
    Assert that all protocols have converged
    Check that Labels are as expected in r1, r2,and r3
    Check ping connectivity between h1 and h2
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # check that r2 peerings are ok
    logger.info("Checking BGP ipv4 vpn summary for r2")
    r2 = tgen.gears["r2"]
    json_file = "{}/{}/ipv4_vpn_summary.json".format(CWD, r2.name)
    expected = json.loads(open(json_file).read())
    test_func = partial(
        topotest.router_json_cmp,
        r2,
        "show bgp ipv4 vpn summary json",
        expected,
    )
    _, result = topotest.run_and_expect(test_func, None, count=20, wait=0.5)
    assertmsg = '"{}" JSON output mismatches'.format(r2.name)
    assert result is None, assertmsg


def test_mpls_setup_ok():
    """
    tests for the r1 to r3 direction: checks for prefix=('172.31.1.0/24','172.31.2.0/24','172.31.3.0/24')
               r2. get label from 'prefix'
    check that r2. show mpls table has an entry with outbound label set to the label from 172.31.1.0/24
               r2. get label from mpls entry
    check that r1: show bgp ipv4 vpn 172.31.1.0/24 has label from r2.mpls entry
    tests for the r3 to r1 direction
               r2. get label from 172.31.0.0/24
    check that r2. show mpls table has an entry with outbound label set that includes the label from 172.31.0.0/24
               r2. get label from mpls entry
    check that r3: show bgp ipv4 vpn 172.31.0.0/24 has label from r2.mpls entry
    check that h1. ping 172.31.1.10 (h2) is ok.
    check that h1. ping 172.31.2.10 (h3) is ok.
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)
    r2 = tgen.gears["r2"]

    # diagnostic
    logger.info("Dumping mplsvpn nexthop table")
    r2.vtysh_cmd("show bgp mplsvpn-nh-label-bind detail")

    vpnv4_checks = {
        "172.31.1.0/24": "r1",
        "172.31.2.0/24": "r1",
        "172.31.3.0/24": "r1",
        "172.31.0.0/24": "r3",
    }
    logger.info(
        "{}, check that 'show bgp ipv4 vpn' and 'show mpls table' are set accordingly on all devices".format(
            r2.name
        )
    )
    check_show_bgp_vpn_ok(r2, vpnv4_checks)

    logger.info("h1, check that ping from h1 to (h2,h3) is ok")
    check_ping("h1", "172.31.1.10", True, 20, 0.5)
    check_ping("h1", "172.31.2.10", True, 20, 0.5)


def test_r3_prefixes_removed():
    """
    Remove BGP redistributed updates from r3.
    Check that the BGP VPN updates from the updates are not present on r2.
    Check that the 'show bgp ipv4 vpn' and 'show mpls table' are ok for 172.31.3.0/24
    Remove the 172.31.3.0/24 update from BGP on r3.
    Check that the BGP VPN updates from r3 are not present on r2.
    Check that the 'show mpls table' entry previously seen disappeared
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)
    r3 = tgen.gears["r3"]
    logger.info("{}, keeping only 172.31.3.0/24 network".format(r3.name))
    r3.vtysh_cmd(
        """
configure terminal
interface r3-eth1 vrf vrf1
 shutdown
!
interface r3-eth2 vrf vrf1
 shutdown\n
"""
    )

    r2 = tgen.gears["r2"]
    logger.info(
        "{}, check that 'show bgp ipv4 vpn' has only 172.31.3.0/24 network from r3".format(
            r2.name
        )
    )

    for prefix in ("172.31.1.0/24", "172.31.2.0/24"):
        test_func = functools.partial(
            check_show_bgp_vpn_prefix_not_found,
            r2,
            "ipv4",
            prefix,
            "444:3",
        )
        success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
        assert success, "{}, vpnv4 update {} still present".format(r2.name, prefix)

    # diagnostic
    logger.info("Dumping mplsvpn nexthop table")
    r2.vtysh_cmd("show bgp mplsvpn-nh-label-bind detail")

    prefix = "172.31.3.0/24"
    logger.info(
        "{}, check that 'show bgp ipv4 vpn' and 'show mpls table' are set accordingly on r2 and on r1".format(
            r2.name
        )
    )
    vpnv4_checks = {
        prefix: "r1",
    }
    label_ip_entries = check_show_bgp_vpn_ok(r2, vpnv4_checks)

    r3 = tgen.gears["r3"]
    logger.info("{}, removing {} network".format(r3.name, prefix))
    r3.vtysh_cmd(
        """
configure terminal
interface r3-eth3 vrf vrf1
 shutdown\n
"""
    )

    r2 = tgen.gears["r2"]
    logger.info(
        "{}, check that 'show bgp ipv4 vpn' has not {} network from r3".format(
            r2.name, prefix
        )
    )
    test_func = functools.partial(
        check_show_bgp_vpn_prefix_not_found,
        r2,
        "ipv4",
        prefix,
        "444:3",
    )
    success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assert success, "{}, vpnv4 update {} still present".format(r2.name, prefix)

    logger.info(
        "{}, check that 'show mpls table {}' is not present".format(
            r2.name, label_ip_entries[prefix]
        )
    )
    test_func = functools.partial(
        check_show_mpls_table_entry_label_not_found, r2, label_ip_entries[prefix]
    )
    success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assert success, "r1, mpls entry with in_label {} still present".format(
        label_ip_entries[prefix]
    )


def test_r3_prefixes_added_back():
    """
    Add back the 172.31.3.0/24 network from r3
    Check on r2 that MPLS switching entry appears when the 1st BGP update is received
    Check the IP connectivity (h1,h2) and (h1,h3)
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)
    r3 = tgen.gears["r3"]
    prefix = "172.31.3.0/24"
    logger.info("{}, restoring the {} network from r3".format(r3.name, prefix))
    r3.vtysh_cmd(
        """
configure terminal
 interface r3-eth3 vrf vrf1
  no shutdown
"""
    )

    r2 = tgen.gears["r2"]
    logger.info(
        "{}, check that 'show bgp ipv4 vpn' has {} network from r3".format(
            r2.name, prefix
        )
    )

    test_func = functools.partial(
        check_show_bgp_vpn_prefix_found,
        r2,
        "ipv4",
        prefix,
        "444:3",
    )
    success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assert success, "{}, vpnv4 update {} not present".format(r2.name, prefix)

    logger.info(
        "{}, check that 'show bgp ipv4 vpn' and 'show mpls table' are set accordingly on r2 and on r1".format(
            r2.name
        )
    )
    vpnv4_checks = {
        prefix: "r1",
    }
    check_show_bgp_vpn_ok(r2, vpnv4_checks)

    r3 = tgen.gears["r3"]
    logger.info(
        "{}, restoring the redistribute connected prefixes from r3".format(r3.name)
    )
    r3.vtysh_cmd(
        """
configure terminal
interface r3-eth1 vrf vrf1
 no shutdown
!
interface r3-eth2 vrf vrf1
 no shutdown
"""
    )

    r2 = tgen.gears["r2"]
    for prefix in ("172.31.1.0/24", "172.31.2.0/24"):
        test_func = functools.partial(
            check_show_bgp_vpn_prefix_found,
            r2,
            "ipv4",
            prefix,
            "444:3",
        )
        success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
        assert success, "{}, vpnv4 update {} not present".format(r2.name, prefix)

    # diagnostic
    logger.info("Dumping mplsvpn nexthop table")
    r2.vtysh_cmd("show bgp mplsvpn-nh-label-bind detail")


def test_unconfigure_nexthop_change_nexthop_self():
    """
    Get the list of labels advertised from r2 to r1
    On r2, disable next-hop-self for 192.0.2.100 neighbor
    Check that the list of labels are not present in 'show mpls table'
    Check that r1 received the prefixes with the original (next-hop,label)
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]
    vpnv4_checks = {
        "172.31.1.0/24": "r1",
        "172.31.2.0/24": "r1",
        "172.31.3.0/24": "r1",
    }
    logger.info(
        "{}, Get the list of labels allocated for prefixes from r3".format(r2.name)
    )
    label_ip_entries = check_show_bgp_vpn_ok(r2, vpnv4_checks)

    logger.info("{}, disable next-hop-self for 192.0.2.100 neighbor".format(r2.name))

    r2.vtysh_cmd(
        """
configure terminal
router bgp 65500
 address-family ipv4 vpn
  no neighbor 192.0.2.100 next-hop-self
"""
    )

    for prefix, label in label_ip_entries.items():
        logger.info(
            "{}, check mpls entry for {} with in_label {} is not present'".format(
                r2.name, prefix, label
            )
        )
        test_func = functools.partial(
            check_show_mpls_table_entry_label_not_found, r2, label
        )
        success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
        assert success, "r1, mpls entry for {} with in_label {} still present".format(
            prefix, label
        )

    r1 = tgen.gears["r1"]
    for prefix, label in label_ip_entries.items():
        test_func = functools.partial(
            check_show_bgp_vpn_prefix_not_found,
            r1,
            "ipv4",
            prefix,
            "444:3",
            label=label,
        )
        success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
        assert success, "{}, mpls vpn update {} label {} is present".format(
            r1.name, prefix, label
        )
    for prefix, label in label_ip_entries.items():
        test_func = functools.partial(
            check_show_bgp_vpn_prefix_found,
            r1,
            "ipv4",
            prefix,
            "444:3",
            nexthop="192.168.1.3",
        )
        success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
        assert success, "{}, mpls vpn update {} label {} is present".format(
            r1.name, prefix, label
        )

    # diagnostic
    logger.info("Dumping mplsvpn nexthop table")
    tgen.gears["r2"].vtysh_cmd("show bgp mplsvpn-nh-label-bind detail")


def test_reconfigure_nexthop_change_nexthop_self():
    """
    Get the list of labels advertised from r2 to r1
    On r2, enable next-hop-self for 192.0.2.100 neighbor
    Check that the list of labels are present in 'show mpls table'
    Check that r1 received the prefixes with the original (next-hop,label)
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]
    logger.info("{}, enable next-hop-self for 192.0.2.100 neighbor".format(r2.name))
    r2.vtysh_cmd(
        """
configure terminal
router bgp 65500
 address-family ipv4 vpn
  neighbor 192.0.2.100 next-hop-self
"""
    )
    vpnv4_checks = {
        "172.31.1.0/24": "r1",
        "172.31.2.0/24": "r1",
        "172.31.3.0/24": "r1",
    }
    logger.info(
        "{}, check that 'show bgp ipv4 vpn' and 'show mpls table' are set accordingly on r2 and on r1".format(
            r2.name
        )
    )
    check_show_bgp_vpn_ok(r2, vpnv4_checks)

    logger.info("h1, check that ping from h1 to (h2,h3) is ok")
    check_ping("h1", "172.31.1.10", True, 20, 0.5)
    check_ping("h1", "172.31.2.10", True, 20, 0.5)
    # diagnostic
    logger.info("Dumping mplsvpn nexthop table")
    r2.vtysh_cmd("show bgp mplsvpn-nh-label-bind detail")


def test_declare_vpn_network_with_different_label():
    """
    declare a vpnv4 network on r3.
    check that a new VPNv4 entry is received on r2.
    Check that the list of labels are present in 'show mpls table'
    Check that r1 received the prefixes with the new (next-hop,label)
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r3 = tgen.gears["r3"]
    logger.info(
        "{}, declare static 33.33.33.33/32 network rd 33:33 label 33".format(r3.name)
    )
    r3.vtysh_cmd(
        """
configure terminal
router bgp 65501
 no bgp network import-check\n"
 address-family ipv4 vpn
  network 33.33.33.33/32 rd 444:3 label 33
"""
    )

    r2 = tgen.gears["r2"]
    vpnv4_entries = {
        "172.31.1.0/24": None,
        "172.31.2.0/24": None,
        "172.31.3.0/24": None,
        "33.33.33.33/32": 33,
    }

    for prefix, label in vpnv4_entries.items():
        test_func = functools.partial(
            check_show_bgp_vpn_prefix_found,
            r2,
            "ipv4",
            prefix,
            "444:3",
            label=label,
            nexthop="192.168.1.3",
        )
        success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
        assert success, "{}, vpnv4 update {}, label {} not present".format(
            r2.name, prefix, label
        )

    vpnv4_checks = {
        "172.31.1.0/24": "r1",
        "172.31.2.0/24": "r1",
        "172.31.3.0/24": "r1",
        "33.33.33.33/32": "r1",
    }
    logger.info(
        "{}, check that 'show bgp ipv4 vpn' and 'show mpls table' are set accordingly on r2 and on r1".format(
            r2.name
        )
    )
    check_show_bgp_vpn_ok(r2, vpnv4_checks)


def test_filter_vpn_network_from_r1():
    """
    Get the list of labels in 'show mpls table'
    filter network from r1
    check that the vpnv4 entry on r2 is not present
    Check that the associated mpls entry is not present
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]

    vpnv4_checks = {
        "172.31.0.0/24": "r3",
    }
    logger.info(
        "{}, check that 'show bgp ipv4 vpn' and 'show mpls table' are set accordingly on r2 and on r3".format(
            r2.name
        )
    )
    label_ip_entries = check_show_bgp_vpn_ok(r2, vpnv4_checks)

    for prefix, label in label_ip_entries.items():
        logger.info("{}, filter prefix {} from r1".format(r2.name, prefix))
        r2.vtysh_cmd(
            """
configure terminal
route-map rmap deny 1
 match ip next-hop address 192.0.2.1
!
router bgp 65500
 address-family ipv4 vpn
  neighbor 192.0.2.100 route-map rmap in
"""
        )
        logger.info("{}, check that prefix {} is not present".format(r2.name, prefix))
        test_func = functools.partial(
            check_show_bgp_vpn_prefix_not_found,
            r2,
            "ipv4",
            "172.31.0.0/24",
            "444:1",
        )
        success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
        assert success, "{}, vpnv4 update {}, is still present".format(r2.name, prefix)

        # diagnostic
        logger.info("Dumping mplsvpn nexthop table")
        r2.vtysh_cmd("show bgp mplsvpn-nh-label-bind detail")

        logger.info(
            "{}, check that show mpls table {} is not present".format(r2.name, label)
        )
        test_func = functools.partial(
            check_show_mpls_table_entry_label_not_found, r2, int(label)
        )
        success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
        assert success, "r1, mpls entry for {} with in_label {} still present".format(
            prefix, label
        )


def test_unfilter_vpn_network_from_r1():
    """
    unfilter network from r1
    check that the vpnv4 entry on r2 is present
    Check that the list of labels are present in 'show mpls table'
    Check that r3 received the prefixes with the new (next-hop,label)
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]
    prefix = "172.31.0.0/24"

    logger.info("{}, filter prefix {} from r1".format(r2.name, prefix))
    r2.vtysh_cmd(
        "configure terminal\nrouter bgp 65500\naddress-family ipv4 vpn\nno neighbor 192.0.2.100 route-map rmap in\n"
    )

    logger.info("{}, check that prefix {} is present".format(r2.name, prefix))
    test_func = functools.partial(
        check_show_bgp_vpn_prefix_found, r2, "ipv4", prefix, "444:1"
    )
    success, _ = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assert success, "{}, vpnv4 update {}, is not present".format(r2.name, prefix)

    vpnv4_checks = {
        "172.31.0.0/24": "r3",
    }
    logger.info(
        "{}, check that 'show bgp ipv4 vpn' and 'show mpls table' are set accordingly on all devices".format(
            r2.name
        )
    )
    check_show_bgp_vpn_ok(r2, vpnv4_checks)

    # diagnostic
    logger.info("Dumping mplsvpn nexthop table")
    r2.vtysh_cmd("show bgp mplsvpn-nh-label-bind detail")


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# SPDX-License-Identifier: ISC
#
# topogen.py
# Library of helper functions for NetDEF Topology Tests
#
# Copyright (c) 2017 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
Topogen (Topology Generator) is an abstraction around Topotest and Mininet to
help reduce boilerplate code and provide a stable interface to build topology
tests on.

Basic usage instructions:

* Define a Topology class with a build method using mininet.topo.Topo.
  See examples/test_template.py.
* Use Topogen inside the build() method with get_topogen.
  e.g. get_topogen(self).
* Start up your topology with: Topogen(YourTopology)
* Initialize the Mininet with your topology with: tgen.start_topology()
* Configure your routers/hosts and start them
* Run your tests / mininet cli.
* After running stop Mininet with: tgen.stop_topology()
"""

import configparser
import grp
import inspect
import json
import logging
import os
import platform
import pwd
import re
import shlex
import subprocess
import sys
from collections import OrderedDict

import lib.topolog as topolog
from lib.micronet import Commander
from lib.micronet_compat import Mininet
from lib.topolog import logger
from munet.testing.util import pause_test

from lib import topotest

CWD = os.path.dirname(os.path.realpath(__file__))

# pylint: disable=C0103
# Global Topogen variable. This is being used to keep the Topogen available on
# all test functions without declaring a test local variable.
global_tgen = None


def get_topogen(topo=None):
    """
    Helper function to retrieve Topogen. Must be called with `topo` when called
    inside the build() method of Topology class.
    """
    if topo is not None:
        global_tgen.topo = topo
    return global_tgen


def set_topogen(tgen):
    "Helper function to set Topogen"
    # pylint: disable=W0603
    global global_tgen
    global_tgen = tgen


def is_string(value):
    """Return True if value is a string."""
    try:
        return isinstance(value, basestring)  # type: ignore
    except NameError:
        return isinstance(value, str)


def get_exabgp_cmd(commander=None):
    """Return the command to use for ExaBGP version >= 4.2.11"""

    if commander is None:
        commander = Commander("exabgp", logger=logging.getLogger("exabgp"))

    def exacmd_version_ok(exacmd):
        logger.debug("checking %s for exabgp version >= 4.2.11", exacmd)
        _, stdout, _ = commander.cmd_status(exacmd + " -v", warn=False)
        m = re.search(r"ExaBGP\s*:\s*((\d+)\.(\d+)(?:\.(\d+))?)", stdout)
        if not m:
            return False
        version = m.group(1)
        if topotest.version_cmp(version, "4.2.11") < 0:
            logging.debug(
                "found exabgp version < 4.2.11 in %s will keep looking", exacmd
            )
            return False
        logger.info("Using ExaBGP version %s in %s", version, exacmd)
        return True

    exacmd = commander.get_exec_path("exabgp")
    if exacmd and exacmd_version_ok(exacmd):
        return exacmd
    py3_path = commander.get_exec_path("python3")
    if py3_path:
        exacmd = py3_path + " -m exabgp"
        if exacmd_version_ok(exacmd):
            return exacmd
    py3_path = commander.get_exec_path("python")
    if py3_path:
        exacmd = py3_path + " -m exabgp"
        if exacmd_version_ok(exacmd):
            return exacmd
    return None


#
# Main class: topology builder
#

# Topogen configuration defaults
tgen_defaults = {
    "verbosity": "info",
    "frrdir": "/usr/lib/frr",
    "routertype": "frr",
    "memleak_path": "",
}


class Topogen(object):
    "A topology test builder helper."

    CONFIG_SECTION = "topogen"

    def __init__(self, topodef, modname="unnamed"):
        """
        Topogen initialization function, takes the following arguments:
        * `cls`: OLD:uthe topology class that is child of mininet.topo or a build function.
        * `topodef`: A dictionary defining the topology, a filename of a json file, or a
          function that will do the same
        * `modname`: module name must be a unique name to identify logs later.
        """
        self.config = None
        self.net = None
        self.gears = {}
        self.routern = 1
        self.switchn = 1
        self.modname = modname
        self.errorsd = {}
        self.errors = ""
        self.peern = 1
        self.cfg_gen = 0
        self.exabgp_cmd = None
        self._init_topo(topodef)

        logger.info("loading topology: {}".format(self.modname))

    # @staticmethod
    # def _mininet_reset():
    #     "Reset the mininet environment"
    #     # Clean up the mininet environment
    #     os.system("mn -c > /dev/null 2>&1")

    def __str__(self):
        return "Topogen()"

    def _init_topo(self, topodef):
        """
        Initialize the topogily provided by the user. The user topology class
        must call get_topogen() during build() to get the topogen object.
        """
        # Set the global variable so the test cases can access it anywhere
        set_topogen(self)

        # Increase host based limits
        topotest.fix_host_limits()

        # Test for MPLS Kernel modules available
        self.hasmpls = False
        if not topotest.module_present("mpls-router"):
            logger.info("MPLS tests will not run (missing mpls-router kernel module)")
        elif not topotest.module_present("mpls-iptunnel"):
            logger.info("MPLS tests will not run (missing mpls-iptunnel kernel module)")
        else:
            self.hasmpls = True

        # Load the default topology configurations
        self._load_config()

        # Create new log directory
        self.logdir = topotest.get_logs_path(topotest.g_pytest_config.option.rundir)
        subprocess.check_call(
            "mkdir -p {0} && chmod 1777 {0}".format(self.logdir), shell=True
        )
        try:
            routertype = self.config.get(self.CONFIG_SECTION, "routertype")
            # Only allow group, if it exist.
            gid = grp.getgrnam(routertype)[2]
            os.chown(self.logdir, 0, gid)
            os.chmod(self.logdir, 0o775)
        except KeyError:
            # Allow anyone, but set the sticky bit to avoid file deletions
            os.chmod(self.logdir, 0o1777)

        # Remove old twisty way of creating sub-classed topology object which has it's
        # build method invoked which calls Topogen methods which then call Topo methods
        # to create a topology within the Topo object, which is then used by
        # Mininet(Micronet) to build the actual topology.
        assert not inspect.isclass(topodef)

        self.net = Mininet(
            rundir=self.logdir,
            pytestconfig=topotest.g_pytest_config,
            logger=topolog.get_logger("mu", log_level="debug"),
        )

        # Adjust the parent namespace
        topotest.fix_netns_limits(self.net)

        # New direct way: Either a dictionary defines the topology or a build function
        # is supplied, or a json filename all of which build the topology by calling
        # Topogen methods which call Mininet(Micronet) methods to create the actual
        # topology.
        if not inspect.isclass(topodef):
            if callable(topodef):
                topodef(self)
                self.net.configure_hosts()
            elif is_string(topodef):
                # topojson imports topogen in one function too,
                # switch away from this use here to the topojson
                # fixutre and remove this case
                from lib.topojson import build_topo_from_json

                with open(topodef, "r") as topof:
                    self.json_topo = json.load(topof)
                build_topo_from_json(self, self.json_topo)
                self.net.configure_hosts()
            elif topodef:
                self.add_topology_from_dict(topodef)

    def add_topology_from_dict(self, topodef):
        keylist = (
            topodef.keys()
            if isinstance(topodef, OrderedDict)
            else sorted(topodef.keys())
        )
        # ---------------------------
        # Create all referenced hosts
        # ---------------------------
        for oname in keylist:
            tup = (topodef[oname],) if is_string(topodef[oname]) else topodef[oname]
            for e in tup:
                desc = e.split(":")
                name = desc[0]
                if name not in self.gears:
                    logging.debug("Adding router: %s", name)
                    self.add_router(name)

        # ------------------------------
        # Create all referenced switches
        # ------------------------------
        for oname in keylist:
            if oname is not None and oname not in self.gears:
                logging.debug("Adding switch: %s", oname)
                self.add_switch(oname)

        # ----------------
        # Create all links
        # ----------------
        for oname in keylist:
            if oname is None:
                continue
            tup = (topodef[oname],) if is_string(topodef[oname]) else topodef[oname]
            for e in tup:
                desc = e.split(":")
                name = desc[0]
                ifname = desc[1] if len(desc) > 1 else None
                sifname = desc[2] if len(desc) > 2 else None
                self.add_link(self.gears[oname], self.gears[name], sifname, ifname)

        self.net.configure_hosts()

    def _load_config(self):
        """
        Loads the configuration file `pytest.ini` located at the root dir of
        topotests.
        """
        self.config = configparser.ConfigParser(tgen_defaults)
        pytestini_path = os.path.join(CWD, "../pytest.ini")
        self.config.read(pytestini_path)

    def add_router(self, name=None, cls=None, **params):
        """
        Adds a new router to the topology. This function has the following
        options:
        * `name`: (optional) select the router name
        * `daemondir`: (optional) custom daemon binary directory
        * `routertype`: (optional) `frr`
        Returns a TopoRouter.
        """
        if cls is None:
            cls = topotest.Router
        if name is None:
            name = "r{}".format(self.routern)
        if name in self.gears:
            raise KeyError("router already exists")

        params["frrdir"] = self.config.get(self.CONFIG_SECTION, "frrdir")
        params["memleak_path"] = self.config.get(self.CONFIG_SECTION, "memleak_path")
        if "routertype" not in params:
            params["routertype"] = self.config.get(self.CONFIG_SECTION, "routertype")

        self.gears[name] = TopoRouter(self, cls, name, **params)
        self.routern += 1
        return self.gears[name]

    def add_switch(self, name=None):
        """
        Adds a new switch to the topology. This function has the following
        options:
        name: (optional) select the switch name
        Returns the switch name and number.
        """
        if name is None:
            name = "s{}".format(self.switchn)
        if name in self.gears:
            raise KeyError("switch already exists")

        self.gears[name] = TopoSwitch(self, name)
        self.switchn += 1
        return self.gears[name]

    def add_exabgp_peer(self, name, ip, defaultRoute):
        """
        Adds a new ExaBGP peer to the topology. This function has the following
        parameters:
        * `ip`: the peer address (e.g. '1.2.3.4/24')
        * `defaultRoute`: the peer default route (e.g. 'via 1.2.3.1')
        """
        if name is None:
            name = "peer{}".format(self.peern)
        if name in self.gears:
            raise KeyError("exabgp peer already exists")

        self.gears[name] = TopoExaBGP(self, name, ip=ip, defaultRoute=defaultRoute)
        self.peern += 1
        return self.gears[name]

    def add_host(self, name, ip, defaultRoute):
        """
        Adds a new host to the topology. This function has the following
        parameters:
        * `ip`: the peer address (e.g. '1.2.3.4/24')
        * `defaultRoute`: the peer default route (e.g. 'via 1.2.3.1')
        """
        if name is None:
            name = "host{}".format(self.peern)
        if name in self.gears:
            raise KeyError("host already exists")

        self.gears[name] = TopoHost(self, name, ip=ip, defaultRoute=defaultRoute)
        self.peern += 1
        return self.gears[name]

    def add_bmp_server(self, name, ip, defaultRoute, port=1789):
        """Add the bmp collector gear"""
        if name in self.gears:
            raise KeyError("The bmp server already exists")

        self.gears[name] = TopoBMPCollector(
            self, name, ip=ip, defaultRoute=defaultRoute, port=port
        )

    def add_link(self, node1, node2, ifname1=None, ifname2=None):
        """
        Creates a connection between node1 and node2. The nodes can be the
        following:
        * TopoGear
          * TopoRouter
          * TopoSwitch
        """
        if not isinstance(node1, TopoGear):
            raise ValueError("invalid node1 type")
        if not isinstance(node2, TopoGear):
            raise ValueError("invalid node2 type")

        if ifname1 is None:
            ifname1 = node1.new_link()
        if ifname2 is None:
            ifname2 = node2.new_link()

        node1.register_link(ifname1, node2, ifname2)
        node2.register_link(ifname2, node1, ifname1)
        self.net.add_link(node1.name, node2.name, ifname1, ifname2)

    def get_gears(self, geartype):
        """
        Returns a dictionary of all gears of type `geartype`.

        Normal usage:
        * Dictionary iteration:
        ```py
        tgen = get_topogen()
        router_dict = tgen.get_gears(TopoRouter)
        for router_name, router in router_dict.items():
            # Do stuff
        ```
        * List iteration:
        ```py
        tgen = get_topogen()
        peer_list = tgen.get_gears(TopoExaBGP).values()
        for peer in peer_list:
            # Do stuff
        ```
        """
        return dict(
            (name, gear)
            for name, gear in self.gears.items()
            if isinstance(gear, geartype)
        )

    def routers(self):
        """
        Returns the router dictionary (key is the router name and value is the
        router object itself).
        """
        return self.get_gears(TopoRouter)

    def exabgp_peers(self):
        """
        Returns the exabgp peer dictionary (key is the peer name and value is
        the peer object itself).
        """
        return self.get_gears(TopoExaBGP)

    def get_bmp_servers(self):
        """
        Retruns the bmp servers dictionnary (the key is the bmp server the
        value is the bmp server object itself).
        """
        return self.get_gears(TopoBMPCollector)

    def start_topology(self):
        """Starts the topology class."""
        logger.info("starting topology: {}".format(self.modname))
        self.net.start()

    def start_router(self, router=None):
        """
        Call the router startRouter method.
        If no router is specified it is called for all registered routers.
        """
        if router is None:
            # pylint: disable=r1704
            # XXX should be hosts?
            for _, router in self.routers().items():
                router.start()
        else:
            if isinstance(router, str):
                router = self.gears[router]

            router.start()

    def stop_topology(self):
        """
        Stops the network topology. This function will call the stop() function
        of all gears before calling the mininet stop function, so they can have
        their oportunity to do a graceful shutdown. stop() is called twice. The
        first is a simple kill with no sleep, the second will sleep if not
        killed and try with a different signal.
        """
        pause = bool(self.net.cfgopt.get_option("--pause-at-end"))
        pause = pause or bool(self.net.cfgopt.get_option("--pause"))
        if pause:
            try:
                pause_test("Before MUNET delete")
            except KeyboardInterrupt:
                print("^C...continuing")
            except Exception as error:
                self.logger.error("\n...continuing after error: %s", error)

        logger.info("stopping topology: {}".format(self.modname))

        errors = ""
        for gear in self.gears.values():
            errors += gear.stop()
        if len(errors) > 0:
            logger.error(
                "Errors found post shutdown - details follow: {}".format(errors)
            )

        try:
            self.net.stop()

        except OSError as error:
            # OSError exception is raised when mininet tries to stop switch
            # though switch is stopped once but mininet tries to stop same
            # switch again, where it ended up with exception

            logger.info(error)
            logger.info("Exception ignored: switch is already stopped")

    def get_exabgp_cmd(self):
        if not self.exabgp_cmd:
            self.exabgp_cmd = get_exabgp_cmd(self.net)
        return self.exabgp_cmd

    def cli(self):
        """
        Interrupt the test and call the command line interface for manual
        inspection. Should be only used on non production code.
        """
        self.net.cli()

    mininet_cli = cli

    def is_memleak_enabled(self):
        "Returns `True` if memory leak report is enable, otherwise `False`."
        # On router failure we can't run the memory leak test
        if self.routers_have_failure():
            return False

        memleak_file = os.environ.get("TOPOTESTS_CHECK_MEMLEAK") or self.config.get(
            self.CONFIG_SECTION, "memleak_path"
        )
        if memleak_file == "" or memleak_file is None:
            return False
        return True

    def report_memory_leaks(self, testname=None):
        "Run memory leak test and reports."
        if not self.is_memleak_enabled():
            return

        # If no name was specified, use the test module name
        if testname is None:
            testname = self.modname

        router_list = self.routers().values()
        for router in router_list:
            router.report_memory_leaks(self.modname)

    def set_error(self, message, code=None):
        "Sets an error message and signal other tests to skip."
        logger.info("setting error msg: %s", message)

        # If no code is defined use a sequential number
        if code is None:
            code = len(self.errorsd)

        self.errorsd[code] = message
        self.errors += "\n{}: {}".format(code, message)

    def has_errors(self):
        "Returns whether errors exist or not."
        return len(self.errorsd) > 0

    def routers_have_failure(self):
        "Runs an assertion to make sure that all routers are running."
        if self.has_errors():
            return True

        errors = ""
        router_list = self.routers().values()
        for router in router_list:
            result = router.check_router_running()
            if result != "":
                errors += result + "\n"

        if errors != "":
            self.set_error(errors, "router_error")
            assert False, errors
            return True
        return False


#
# Topology gears (equipment)
#


class TopoGear(object):
    "Abstract class for type checking"

    def __init__(self, tgen, name, **params):
        self.tgen = tgen
        self.name = name
        self.params = params
        self.links = {}
        self.linkn = 0

        # Would be nice for this to point at the gears log directory rather than the
        # test's.
        self.logdir = tgen.logdir
        self.gearlogdir = None

    def __str__(self):
        links = ""
        for myif, dest in self.links.items():
            _, destif = dest
            if links != "":
                links += ","
            links += '"{}"<->"{}"'.format(myif, destif)

        return 'TopoGear<name="{}",links=[{}]>'.format(self.name, links)

    @property
    def net(self):
        return self.tgen.net[self.name]

    def start(self):
        "Basic start function that just reports equipment start"
        logger.info('starting "{}"'.format(self.name))

    def stop(self, wait=True, assertOnError=True):
        "Basic stop function that just reports equipment stop"
        logger.info('"{}" base stop called'.format(self.name))
        return ""

    def cmd(self, command, **kwargs):
        """
        Runs the provided command string in the router and returns a string
        with the response.
        """
        return self.net.cmd_legacy(command, **kwargs)

    def cmd_raises(self, command, **kwargs):
        """
        Runs the provided command string in the router and returns a string
        with the response. Raise an exception on any error.
        """
        return self.net.cmd_raises(command, **kwargs)

    run = cmd

    def popen(self, *params, **kwargs):
        """
        Creates a pipe with the given command. Same args as python Popen.
        If `command` is a string then will be invoked with shell, otherwise
        `command` is a list and will be invoked w/o shell. Returns a popen object.
        """
        return self.net.popen(*params, **kwargs)

    def add_link(self, node, myif=None, nodeif=None):
        """
        Creates a link (connection) between myself and the specified node.
        Interfaces name can be speficied with:
        myif: the interface name that will be created in this node
        nodeif: the target interface name that will be created on the remote node.
        """
        self.tgen.add_link(self, node, myif, nodeif)

    def link_enable(self, myif, enabled=True, netns=None):
        """
        Set this node interface administrative state.
        myif: this node interface name
        enabled: whether we should enable or disable the interface
        """
        if myif not in self.links.keys():
            raise KeyError("interface doesn't exists")

        if enabled is True:
            operation = "up"
        else:
            operation = "down"

        logger.info(
            'setting node "{}" link "{}" to state "{}"'.format(
                self.name, myif, operation
            )
        )
        extract = ""
        if netns is not None:
            extract = "ip netns exec {} ".format(netns)

        return self.run("{}ip link set dev {} {}".format(extract, myif, operation))

    def peer_link_enable(self, myif, enabled=True, netns=None):
        """
        Set the peer interface administrative state.
        myif: this node interface name
        enabled: whether we should enable or disable the interface

        NOTE: this is used to simulate a link down on this node, since when the
        peer disables their interface our interface status changes to no link.
        """
        if myif not in self.links.keys():
            raise KeyError("interface doesn't exists")

        node, nodeif = self.links[myif]
        node.link_enable(nodeif, enabled, netns)

    def new_link(self):
        """
        Generates a new unique link name.

        NOTE: This function should only be called by Topogen.
        """
        ifname = "{}-eth{}".format(self.name, self.linkn)
        self.linkn += 1
        return ifname

    def register_link(self, myif, node, nodeif):
        """
        Register link between this node interface and outside node.

        NOTE: This function should only be called by Topogen.
        """
        if myif in self.links.keys():
            raise KeyError("interface already exists")

        self.links[myif] = (node, nodeif)

    def _setup_tmpdir(self):
        topotest.setup_node_tmpdir(self.logdir, self.name)
        self.gearlogdir = "{}/{}".format(self.logdir, self.name)
        return "{}/{}.log".format(self.logdir, self.name)


class TopoRouter(TopoGear):
    """
    Router abstraction.
    """

    # The default required directories by FRR
    PRIVATE_DIRS = [
        "/etc/frr",
        "/etc/snmp",
        "/var/run/frr",
        "/var/lib/frr",
        "/var/log",
    ]

    # Router Daemon enumeration definition.
    RD_FRR = 0  # not a daemon, but use to setup unified configs
    RD_ZEBRA = 1
    RD_RIP = 2
    RD_RIPNG = 3
    RD_OSPF = 4
    RD_OSPF6 = 5
    RD_ISIS = 6
    RD_BGP = 7
    RD_LDP = 8
    RD_PIM = 9
    RD_EIGRP = 10
    RD_NHRP = 11
    RD_STATIC = 12
    RD_BFD = 13
    RD_SHARP = 14
    RD_BABEL = 15
    RD_PBRD = 16
    RD_PATH = 17
    RD_SNMP = 18
    RD_PIM6 = 19
    RD_MGMTD = 20
    RD_TRAP = 21
    RD_FPM_LISTENER = 22
    RD = {
        RD_FRR: "frr",
        RD_ZEBRA: "zebra",
        RD_RIP: "ripd",
        RD_RIPNG: "ripngd",
        RD_OSPF: "ospfd",
        RD_OSPF6: "ospf6d",
        RD_ISIS: "isisd",
        RD_BGP: "bgpd",
        RD_PIM: "pimd",
        RD_PIM6: "pim6d",
        RD_LDP: "ldpd",
        RD_EIGRP: "eigrpd",
        RD_NHRP: "nhrpd",
        RD_STATIC: "staticd",
        RD_BFD: "bfdd",
        RD_SHARP: "sharpd",
        RD_BABEL: "babeld",
        RD_PBRD: "pbrd",
        RD_PATH: "pathd",
        RD_SNMP: "snmpd",
        RD_MGMTD: "mgmtd",
        RD_TRAP: "snmptrapd",
        RD_FPM_LISTENER: "fpm_listener",
    }

    def __init__(self, tgen, cls, name, **params):
        """
        The constructor has the following parameters:
        * tgen: Topogen object
        * cls: router class that will be used to instantiate
        * name: router name
        * daemondir: daemon binary directory
        * routertype: 'frr'
        """
        super(TopoRouter, self).__init__(tgen, name, **params)
        self.routertype = params.get("routertype", "frr")
        if "private_mounts" not in params:
            params["private_mounts"] = self.PRIVATE_DIRS

        # Propagate the router log directory
        logfile = self._setup_tmpdir()
        params["logdir"] = self.logdir

        self.logger = topolog.get_logger(name, log_level="debug", target=logfile)
        params["logger"] = self.logger
        tgen.net.add_host(self.name, cls=cls, **params)
        topotest.fix_netns_limits(tgen.net[name])

        # Mount gear log directory on a common path
        self.net.bind_mount(self.gearlogdir, "/tmp/gearlogdir")

        # Ensure pid file
        with open(os.path.join(self.logdir, self.name + ".pid"), "w") as f:
            f.write(str(self.net.pid) + "\n")

    def __str__(self):
        gear = super(TopoRouter, self).__str__()
        gear += " TopoRouter<>"
        return gear

    def use_netns_vrf(self):
        """
        Use netns as VRF backend.
        """
        self.net.useNetnsVRF()

    def check_capability(self, daemon, param):
        """
        Checks a capability daemon against an argument option
        Return True if capability available. False otherwise
        """
        daemonstr = self.RD.get(daemon)
        self.logger.info('check capability {} for "{}"'.format(param, daemonstr))
        return self.net.checkCapability(daemonstr, param)

    def load_frr_config(self, source, daemons=None):
        """
        Loads the unified configuration file source
        Start the daemons in the list
        If daemons is None, try to infer daemons from the config file
        `daemons` is a tuple (daemon, param) of daemons to start, e.g.:
        (TopoRouter.RD_ZEBRA, "-s 90000000").
        """
        source_path = self.load_config(self.RD_FRR, source)
        if not daemons:
            # Always add zebra
            self.load_config(self.RD_ZEBRA, "")
            for daemon in self.RD:
                # This will not work for all daemons
                daemonstr = self.RD.get(daemon).rstrip("d")
                if daemonstr == "path":
                    grep_cmd = "grep 'candidate-path' {}".format(source_path)
                else:
                    grep_cmd = "grep -w '{}' {}".format(daemonstr, source_path)
                result = self.run(grep_cmd, warn=False).strip()
                if result:
                    self.load_config(daemon, "")
                    if daemonstr == "ospf":
                        grep_cmd = "grep -E 'router ospf ([0-9]+*)' {} | grep -o -E '([0-9]*)'".format(
                            source_path
                        )
                        result = self.run(grep_cmd, warn=False)
                        if result:  # instances
                            instances = result.split("\n")
                            for inst in instances:
                                if inst != "":
                                    self.load_config(daemon, "", None, inst)

        else:
            for item in daemons:
                daemon, param = item
                self.load_config(daemon, "", param)

    def load_config(self, daemon, source=None, param=None, instance=None):
        """Loads daemon configuration from the specified source
        Possible daemon values are: TopoRouter.RD_ZEBRA, TopoRouter.RD_RIP,
        TopoRouter.RD_RIPNG, TopoRouter.RD_OSPF, TopoRouter.RD_OSPF6,
        TopoRouter.RD_ISIS, TopoRouter.RD_BGP, TopoRouter.RD_LDP,
        TopoRouter.RD_PIM, TopoRouter.RD_PIM6, TopoRouter.RD_PBR,
        TopoRouter.RD_SNMP, TopoRouter.RD_MGMTD, TopoRouter.RD_TRAP,
        TopoRouter.RD_FPM_LISTENER.

        Possible `source` values are `None` for an empty config file, a path name which is
        used directly, or a file name with no path components which is first looked for
        directly and then looked for under a sub-directory named after router.

        This API unfortunately allows for source to not exist for any and
        all routers.
        """
        daemonstr = self.RD.get(daemon)
        self.logger.debug('loading "{}" configuration: {}'.format(daemonstr, source))
        return self.net.loadConf(daemonstr, source, param, instance)

    def check_router_running(self):
        """
        Run a series of checks and returns a status string.
        """
        self.logger.info("checking if daemons are running")
        return self.net.checkRouterRunning()

    def start(self):
        """
        Start router:
        * Load modules
        * Clean up files
        * Configure interfaces
        * Start daemons (e.g. FRR)
        * Configure daemon logging files
        """

        nrouter = self.net
        result = nrouter.startRouter(self.tgen)

        # Enable command logging

        # Enable all daemon command logging, logging files
        # and set them to the start dir.
        for daemon, enabled in nrouter.daemons.items():
            if (
                enabled
                and daemon != "snmpd"
                and daemon != "snmptrapd"
                and daemon != "fpm_listener"
            ):
                self.vtysh_cmd(
                    "\n".join(
                        [
                            "clear log cmdline-targets",
                            "conf t",
                            "log file {}.log debug".format(daemon),
                            "log commands",
                            "log timestamp precision 6",
                        ]
                    ),
                    daemon=daemon,
                )

        if result != "":
            self.tgen.set_error(result)
        elif nrouter.daemons["ldpd"] == 1 or nrouter.daemons["pathd"] == 1:
            # Enable MPLS processing on all interfaces.
            for interface in self.links:
                topotest.sysctl_assure(
                    nrouter, "net.mpls.conf.{}.input".format(interface), 1
                )

        return result

    def stop(self):
        """
        Stop router cleanly:
        * Signal daemons twice, once with SIGTERM, then with SIGKILL.
        """
        self.logger.debug("stopping (no assert)")
        return self.net.stopRouter(False)

    def startDaemons(self, daemons):
        """
        Start Daemons: to start specific daemon(user defined daemon only)
        * Start daemons (e.g. FRR)
        * Configure daemon logging files
        """
        self.logger.debug("starting")
        nrouter = self.net
        result = nrouter.startRouterDaemons(daemons)

        if daemons is None:
            daemons = nrouter.daemons.keys()

        # Enable all daemon command logging, logging files
        # and set them to the start dir.
        for daemon in daemons:
            enabled = nrouter.daemons[daemon]
            if enabled and daemon != "snmpd" and daemon != "fpm_listener":
                self.vtysh_cmd(
                    "\n".join(
                        [
                            "clear log cmdline-targets",
                            "conf t",
                            "log file {}.log debug".format(daemon),
                            "log commands",
                            "log timestamp precision 6",
                        ]
                    ),
                    daemon=daemon,
                )

        if result != "":
            self.tgen.set_error(result)

        return result

    def killDaemons(self, daemons, wait=True, assertOnError=True):
        """
        Kill specific daemon(user defined daemon only)
        forcefully using SIGKILL
        """
        self.logger.debug("Killing daemons using SIGKILL..")
        return self.net.killRouterDaemons(daemons, wait, assertOnError)

    def vtysh_cmd(self, command, isjson=False, daemon=None):
        """
        Runs the provided command string in the vty shell and returns a string
        with the response.

        This function also accepts multiple commands, but this mode does not
        return output for each command. See vtysh_multicmd() for more details.
        """
        # Detect multi line commands
        if command.find("\n") != -1:
            return self.vtysh_multicmd(command, daemon=daemon)

        dparam = ""
        if daemon is not None:
            dparam += "-d {}".format(daemon)

        vtysh_command = "vtysh {} -c {} 2>/dev/null".format(
            dparam, shlex.quote(command)
        )

        self.logger.debug("vtysh command => {}".format(shlex.quote(command)))
        output = self.run(vtysh_command)

        dbgout = output.strip()
        if dbgout:
            if "\n" in dbgout:
                dbgout = dbgout.replace("\n", "\n\t")
                self.logger.debug("vtysh result:\n\t{}".format(dbgout))
            else:
                self.logger.debug('vtysh result: "{}"'.format(dbgout))

        if isjson is False:
            return output

        try:
            return json.loads(output)
        except ValueError as error:
            logger.warning(
                "vtysh_cmd: %s: failed to convert json output: %s: %s",
                self.name,
                str(output),
                str(error),
            )
            return {}

    def vtysh_multicmd(self, commands, pretty_output=True, daemon=None):
        """
        Runs the provided commands in the vty shell and return the result of
        execution.

        pretty_output: defines how the return value will be presented. When
        True it will show the command as they were executed in the vty shell,
        otherwise it will only show lines that failed.
        """
        # Prepare the temporary file that will hold the commands
        fname = topotest.get_file(commands)

        dparam = ""
        if daemon is not None:
            dparam += "-d {}".format(daemon)

        # Run the commands and delete the temporary file
        if pretty_output:
            vtysh_command = "vtysh {} < {}".format(dparam, fname)
        else:
            vtysh_command = "vtysh {} -f {}".format(dparam, fname)

        dbgcmds = commands if is_string(commands) else "\n".join(commands)
        dbgcmds = "\t" + dbgcmds.replace("\n", "\n\t")
        self.logger.debug("vtysh command => FILE:\n{}".format(dbgcmds))

        res = self.run(vtysh_command)
        os.unlink(fname)

        dbgres = res.strip()
        if dbgres:
            if "\n" in dbgres:
                dbgres = dbgres.replace("\n", "\n\t")
                self.logger.debug("vtysh result:\n\t{}".format(dbgres))
            else:
                self.logger.debug('vtysh result: "{}"'.format(dbgres))
        return res

    def report_memory_leaks(self, testname):
        """
        Runs the router memory leak check test. Has the following parameter:
        testname: the test file name for identification

        NOTE: to run this you must have the environment variable
        TOPOTESTS_CHECK_MEMLEAK set or memleak_path configured in `pytest.ini`.
        """
        memleak_file = (
            os.environ.get("TOPOTESTS_CHECK_MEMLEAK") or self.params["memleak_path"]
        )
        if memleak_file == "" or memleak_file is None:
            return

        self.stop()

        self.logger.info("running memory leak report")
        self.net.report_memory_leaks(memleak_file, testname)

    def version_info(self):
        "Get equipment information from 'show version'."
        output = self.vtysh_cmd("show version").split("\n")[0]
        columns = topotest.normalize_text(output).split(" ")
        try:
            return {
                "type": columns[0],
                "version": columns[1],
            }
        except IndexError:
            return {
                "type": None,
                "version": None,
            }

    def has_version(self, cmpop, version):
        """
        Compares router version using operation `cmpop` with `version`.
        Valid `cmpop` values:
        * `>=`: has the same version or greater
        * '>': has greater version
        * '=': has the same version
        * '<': has a lesser version
        * '<=': has the same version or lesser

        Usage example: router.has_version('>', '1.0')
        """
        return self.net.checkRouterVersion(cmpop, version)

    def has_type(self, rtype):
        """
        Compares router type with `rtype`. Returns `True` if the type matches,
        otherwise `false`.
        """
        curtype = self.version_info()["type"]
        return rtype == curtype

    def has_mpls(self):
        return self.net.hasmpls


class TopoSwitch(TopoGear):
    """
    Switch abstraction. Has the following properties:
    * cls: switch class that will be used to instantiate
    * name: switch name
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, tgen, name, **params):
        logger = topolog.get_logger(name, log_level="debug")
        super(TopoSwitch, self).__init__(tgen, name, **params)
        tgen.net.add_switch(name, logger=logger)

    def __str__(self):
        gear = super(TopoSwitch, self).__str__()
        gear += " TopoSwitch<>"
        return gear


class TopoHost(TopoGear):
    "Host abstraction."
    # pylint: disable=too-few-public-methods

    def __init__(self, tgen, name, **params):
        """
        Mininet has the following known `params` for hosts:
        * `ip`: the IP address (string) for the host interface
        * `defaultRoute`: the default route that will be installed
          (e.g. 'via 10.0.0.1')
        * `private_mounts`: directories that will be mounted on a different domain
          (e.g. '/etc/important_dir').
        """
        super(TopoHost, self).__init__(tgen, name, **params)

        # Propagate the router log directory
        logfile = self._setup_tmpdir()
        params["logdir"] = self.logdir

        # Odd to have 2 logfiles for each host
        self.logger = topolog.get_logger(name, log_level="debug", target=logfile)
        params["logger"] = self.logger
        tgen.net.add_host(name, **params)
        topotest.fix_netns_limits(tgen.net[name])

        # Mount gear log directory on a common path
        self.net.bind_mount(self.gearlogdir, "/tmp/gearlogdir")

    def __str__(self):
        gear = super(TopoHost, self).__str__()
        gear += ' TopoHost<ip="{}",defaultRoute="{}",private_mounts="{}">'.format(
            self.params["ip"],
            self.params["defaultRoute"],
            str(self.params["private_mounts"]),
        )
        return gear


class TopoExaBGP(TopoHost):
    "ExaBGP peer abstraction."
    # pylint: disable=too-few-public-methods

    PRIVATE_DIRS = [
        "/etc/exabgp",
        "/var/run/exabgp",
        "/var/log",
    ]

    def __init__(self, tgen, name, **params):
        """
        ExaBGP usually uses the following parameters:
        * `ip`: the IP address (string) for the host interface
        * `defaultRoute`: the default route that will be installed
          (e.g. 'via 10.0.0.1')

        Note: the different between a host and a ExaBGP peer is that this class
        has a private_mounts already defined and contains functions to handle
        ExaBGP things.
        """
        params["private_mounts"] = self.PRIVATE_DIRS
        super(TopoExaBGP, self).__init__(tgen, name, **params)

    def __str__(self):
        gear = super(TopoExaBGP, self).__str__()
        gear += " TopoExaBGP<>".format()
        return gear

    def start(self, peer_dir, env_file=None):
        """
        Start running ExaBGP daemon:
        * Copy all peer* folder contents into /etc/exabgp
        * Copy exabgp env file if specified
        * Make all python files runnable
        * Run ExaBGP with env file `env_file` and configuration peer*/exabgp.cfg
        """
        exacmd = self.tgen.get_exabgp_cmd()
        assert exacmd, "Can't find a usable ExaBGP (must be version >= 4.2.11)"

        self.run("mkdir -p /etc/exabgp")
        self.run("chmod 755 /etc/exabgp")
        self.run("cp {}/exa-* /etc/exabgp/".format(CWD))
        self.run("cp {}/* /etc/exabgp/".format(peer_dir))
        if env_file is not None:
            self.run("cp {} /etc/exabgp/exabgp.env".format(env_file))
        self.run("chmod 644 /etc/exabgp/*")
        self.run("chmod a+x /etc/exabgp/*.py")
        self.run("chown -R exabgp:exabgp /etc/exabgp")
        self.run("[ -p /var/run/exabgp.in ] || mkfifo /var/run/exabgp.in")
        self.run("[ -p /var/run/exabgp.out ] || mkfifo /var/run/exabgp.out")
        self.run("chown exabgp:exabgp /var/run/exabgp.{in,out}")
        self.run("chmod 600 /var/run/exabgp.{in,out}")

        log_dir = os.path.join(self.logdir, self.name)
        self.run("chmod 777 {}".format(log_dir))

        log_file = os.path.join(log_dir, "exabgp.log")

        env_cmd = "env exabgp.log.level=INFO "
        env_cmd += "exabgp.log.destination={} ".format(log_file)

        output = self.run(
            env_cmd + exacmd + " -e /etc/exabgp/exabgp.env /etc/exabgp/exabgp.cfg "
        )
        if output is None or len(output) == 0:
            output = "<none>"

        logger.info("{} exabgp started, output={}".format(self.name, output))

    def stop(self, wait=True, assertOnError=True):
        "Stop ExaBGP peer and kill the daemon"
        self.run("kill `cat /var/run/exabgp/exabgp.pid`")
        return ""


class TopoBMPCollector(TopoHost):
    PRIVATE_DIRS = [
        "/var/log",
    ]

    def __init__(self, tgen, name, **params):
        params["private_mounts"] = self.PRIVATE_DIRS
        self.port = params["port"]
        self.ip = params["ip"]
        super(TopoBMPCollector, self).__init__(tgen, name, **params)

    def __str__(self):
        gear = super(TopoBMPCollector, self).__str__()
        gear += " TopoBMPCollector<>".format()
        return gear

    def start(self, log_file=None):
        log_dir = os.path.join(self.logdir, self.name)
        self.run("chmod 777 {}".format(log_dir))

        log_err = os.path.join(log_dir, "bmpserver.log")

        log_arg = "-l {}".format(log_file) if log_file else ""
        self.pid_file = os.path.join(log_dir, "bmpserver.pid")

        with open(log_err, "w") as err:
            self.run(
                "{}/bmp_collector/bmpserver.py -a {} -p {} -r {} {}&".format(
                    CWD, self.ip, self.port, self.pid_file, log_arg
                ),
                stdout=None,
                stderr=err,
            )

    def stop(self):
        self.run(f"kill $(cat {self.pid_file}")
        return ""


#
# Diagnostic function
#


# Disable linter branch warning. It is expected to have these here.
# pylint: disable=R0912
def diagnose_env_linux(rundir):
    """
    Run diagnostics in the running environment. Returns `True` when everything
    is ok, otherwise `False`.
    """
    ret = True

    # Load configuration
    config = configparser.ConfigParser(defaults=tgen_defaults)
    pytestini_path = os.path.join(CWD, "../pytest.ini")
    config.read(pytestini_path)

    # Test log path exists before installing handler.
    os.system("mkdir -p " + rundir)
    # Log diagnostics to file so it can be examined later.
    fhandler = logging.FileHandler(filename="{}/diagnostics.txt".format(rundir))
    fhandler.setLevel(logging.DEBUG)
    fhandler.setFormatter(logging.Formatter(fmt=topolog.FORMAT))
    logger.addHandler(fhandler)

    logger.info("Running environment diagnostics")

    # Assert that we are running as root
    if os.getuid() != 0:
        logger.error("you must run topotest as root")
        ret = False

    # Assert that we have mininet
    # if os.system("which mn >/dev/null 2>/dev/null") != 0:
    #     logger.error("could not find mininet binary (mininet is not installed)")
    #     ret = False

    # Assert that we have iproute installed
    if os.system("which ip >/dev/null 2>/dev/null") != 0:
        logger.error("could not find ip binary (iproute is not installed)")
        ret = False

    # Assert that we have gdb installed
    if os.system("which gdb >/dev/null 2>/dev/null") != 0:
        logger.error("could not find gdb binary (gdb is not installed)")
        ret = False

    # Assert that FRR utilities exist
    frrdir = config.get("topogen", "frrdir")
    if not os.path.isdir(frrdir):
        logger.error("could not find {} directory".format(frrdir))
        ret = False
    else:
        try:
            pwd.getpwnam("frr")[2]
        except KeyError:
            logger.warning('could not find "frr" user')

        try:
            grp.getgrnam("frr")[2]
        except KeyError:
            logger.warning('could not find "frr" group')

        try:
            if "frr" not in grp.getgrnam("frrvty").gr_mem:
                logger.error(
                    '"frr" user and group exist, but user is not under "frrvty"'
                )
        except KeyError:
            logger.warning('could not find "frrvty" group')

        for fname in [
            "zebra",
            "ospfd",
            "ospf6d",
            "bgpd",
            "ripd",
            "ripngd",
            "isisd",
            "pimd",
            "pim6d",
            "ldpd",
            "pbrd",
            "mgmtd",
        ]:
            path = os.path.join(frrdir, fname)
            if not os.path.isfile(path):
                # LDPd is an exception
                if fname == "ldpd":
                    logger.info(
                        "could not find {} in {}".format(fname, frrdir)
                        + "(LDPd tests will not run)"
                    )
                    continue

                logger.error("could not find {} in {}".format(fname, frrdir))
                ret = False
            else:
                if fname != "zebra" or fname != "mgmtd":
                    continue

                os.system("{} -v 2>&1 >{}/frr_mgmtd.txt".format(path, rundir))
                os.system("{} -v 2>&1 >{}/frr_zebra.txt".format(path, rundir))

    # Test MPLS availability
    krel = platform.release()
    if topotest.version_cmp(krel, "4.5") < 0:
        logger.info(
            'LDPd tests will not run (have kernel "{}", but it requires 4.5)'.format(
                krel
            )
        )

    # Test for MPLS Kernel modules available
    if not topotest.module_present("mpls-router", load=False) != 0:
        logger.info("LDPd tests will not run (missing mpls-router kernel module)")
    if not topotest.module_present("mpls-iptunnel", load=False) != 0:
        logger.info("LDPd tests will not run (missing mpls-iptunnel kernel module)")

    if not get_exabgp_cmd():
        logger.warning("Failed to find exabgp >= 4.2.11")

    logger.removeHandler(fhandler)
    fhandler.close()

    return ret


def diagnose_env_freebsd():
    return True


def diagnose_env(rundir):
    if sys.platform.startswith("linux"):
        return diagnose_env_linux(rundir)
    elif sys.platform.startswith("freebsd"):
        return diagnose_env_freebsd()

    return False



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2019 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

import sys
import time
import pytest
import os

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen

# Required to instantiate the topology builder class.
from lib.common_config import (
    start_topology,
    write_test_header,
    write_test_footer,
    verify_rib,
    create_route_maps,
    create_static_routes,
    create_prefix_lists,
    check_address_types,
    reset_config_on_routers,
)
from lib.topolog import logger
from lib.bgp import (
    verify_bgp_convergence,
    create_router_bgp,
    verify_bgp_attributes,
)
from lib.topojson import build_config_from_json

pytestmark = [pytest.mark.bgpd, pytest.mark.staticd]


#################################
# TOPOLOGY
#################################
"""

                    +-------+
           +------- |  R2   |
          |         +-------+
         |               |
     +-------+           |
     |  R1   |           |
     +-------+           |
        |                |
        |           +-------+          +-------+
        +---------- |  R3   |----------|  R4   |
                    +-------+          +-------+

"""

#################################
# TEST SUMMARY
#################################
"""
Following tests are covered to test route-map functionality:
TC_34:
    Verify if route-maps is applied in both inbound and
    outbound direction to same neighbor/interface.
TC_36:
    Test permit/deny statements operation in route-maps with a
    permutation and combination of permit/deny in prefix-lists
TC_35:
    Test multiple sequence numbers in a single route-map for different
    match/set clauses.
TC_37:
    Test add/remove route-maps with multiple set
    clauses and without any match statement.(Set only)
TC_38:
    Test add/remove route-maps with multiple match
    clauses and without any set statement.(Match only)
"""

# Global variables
bgp_convergence = False
BGP_CONVERGENCE = False
ADDR_TYPES = check_address_types()

# Global variables
bgp_convergence = False
NETWORK = {"ipv4": ["11.0.20.1/32", "20.0.20.1/32"], "ipv6": ["1::1/128", "2::1/128"]}
MASK = {"ipv4": "32", "ipv6": "128"}
NEXT_HOP = {"ipv4": "10.0.0.2", "ipv6": "fd00::2"}
ADDR_TYPES = check_address_types()


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """
    global ADDR_TYPES
    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/bgp_route_map_topo1.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # Checking BGP convergence
    global bgp_convergence

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Api call verify whether BGP is converged
    bgp_convergence = verify_bgp_convergence(tgen, topo)
    assert bgp_convergence is True, "setup_module :Failed \n Error:" " {}".format(
        bgp_convergence
    )

    logger.info("Running setup_module() done")


def teardown_module():
    """
    Teardown the pytest environment

    * `mod`: module name
    """

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


def test_route_map_inbound_outbound_same_neighbor_p0(request):
    """
    TC_34:
    Verify if route-maps is applied in both inbound and
    outbound direction to same neighbor/interface.
    """

    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    for adt in ADDR_TYPES:

        # Create Static routes
        input_dict = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[adt][0],
                        "no_of_ip": 9,
                        "next_hop": NEXT_HOP[adt],
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Api call to redistribute static routes
        input_dict_1 = {
            "r1": {
                "bgp": {
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                    },
                }
            }
        }

        result = create_router_bgp(tgen, topo, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        input_dict_2 = {
            "r4": {
                "static_routes": [
                    {
                        "network": NETWORK[adt][1],
                        "no_of_ip": 9,
                        "next_hop": NEXT_HOP[adt],
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict_2)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Api call to redistribute static routes
        input_dict_5 = {
            "r1": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_5)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        input_dict_2 = {
            "r3": {
                "prefix_lists": {
                    "ipv4": {
                        "pf_list_1_ipv4": [
                            {
                                "seqid": 10,
                                "action": "permit",
                                "network": NETWORK["ipv4"][0],
                            }
                        ],
                        "pf_list_2_ipv4": [
                            {
                                "seqid": 10,
                                "action": "permit",
                                "network": NETWORK["ipv4"][1],
                            }
                        ],
                    },
                    "ipv6": {
                        "pf_list_1_ipv6": [
                            {
                                "seqid": 100,
                                "action": "permit",
                                "network": NETWORK["ipv6"][0],
                            }
                        ],
                        "pf_list_2_ipv6": [
                            {
                                "seqid": 100,
                                "action": "permit",
                                "network": NETWORK["ipv6"][1],
                            }
                        ],
                    },
                }
            }
        }
        result = create_prefix_lists(tgen, input_dict_2)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Create route map
        for addr_type in ADDR_TYPES:
            input_dict_6 = {
                "r3": {
                    "route_maps": {
                        "rmap_match_tag_1_{}".format(addr_type): [
                            {
                                "action": "deny",
                                "match": {
                                    addr_type: {
                                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                                    }
                                },
                            }
                        ],
                        "rmap_match_tag_2_{}".format(addr_type): [
                            {
                                "action": "permit",
                                "match": {
                                    addr_type: {
                                        "prefix_lists": "pf_list_2_{}".format(addr_type)
                                    }
                                },
                            }
                        ],
                    }
                }
            }
            result = create_route_maps(tgen, input_dict_6)
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )

        # Configure neighbor for route map
        input_dict_7 = {
            "r3": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "neighbor": {
                                    "r4": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_tag_1_ipv4",
                                                        "direction": "in",
                                                    },
                                                    {
                                                        "name": "rmap_match_tag_1_ipv4",
                                                        "direction": "out",
                                                    },
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r4": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_tag_1_ipv6",
                                                        "direction": "in",
                                                    },
                                                    {
                                                        "name": "rmap_match_tag_1_ipv6",
                                                        "direction": "out",
                                                    },
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        },
                    }
                }
            }
        }

        result = create_router_bgp(tgen, topo, input_dict_7)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    for adt in ADDR_TYPES:
        # Verifying RIB routes
        dut = "r3"
        protocol = "bgp"
        input_dict_2 = {
            "r4": {
                "static_routes": [
                    {
                        "network": [NETWORK[adt][1]],
                        "no_of_ip": 9,
                        "next_hop": NEXT_HOP[adt],
                    }
                ]
            }
        }

        result = verify_rib(
            tgen, adt, dut, input_dict_2, protocol=protocol, expected=False
        )
        assert result is not True, (
            "Testcase {} : Failed \n "
            "Expected: Routes should not be present in {} BGP RIB \n "
            "Found: {}".format(tc_name, dut, result)
        )

        # Verifying RIB routes
        dut = "r4"
        input_dict = {
            "r1": {
                "static_routes": [
                    {
                        "network": [NETWORK[adt][0]],
                        "no_of_ip": 9,
                        "next_hop": NEXT_HOP[adt],
                    }
                ]
            }
        }
        result = verify_rib(
            tgen, adt, dut, input_dict, protocol=protocol, expected=False
        )
        assert result is not True, (
            "Testcase {} : Failed \n "
            "Expected: Routes should not be present in {} FIB \n "
            "Found: {}".format(tc_name, dut, result)
        )

    write_test_footer(tc_name)


@pytest.mark.parametrize(
    "prefix_action, rmap_action",
    [("permit", "permit"), ("permit", "deny"), ("deny", "permit"), ("deny", "deny")],
)
def test_route_map_with_action_values_combination_of_prefix_action_p0(
    request, prefix_action, rmap_action
):
    """
    TC_36:
    Test permit/deny statements operation in route-maps with a permutation and
    combination of permit/deny in prefix-lists
    """
    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    for adt in ADDR_TYPES:
        # Create Static routes
        input_dict = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[adt][0],
                        "no_of_ip": 9,
                        "next_hop": NEXT_HOP[adt],
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Api call to redistribute static routes
        input_dict_1 = {
            "r1": {
                "bgp": {
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                    },
                }
            }
        }

        result = create_router_bgp(tgen, topo, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Permit in perfix list and route-map
        input_dict_2 = {
            "r3": {
                "prefix_lists": {
                    "ipv4": {
                        "pf_list_1_ipv4": [
                            {"seqid": 10, "network": "any", "action": prefix_action}
                        ]
                    },
                    "ipv6": {
                        "pf_list_1_ipv6": [
                            {"seqid": 100, "network": "any", "action": prefix_action}
                        ]
                    },
                }
            }
        }
        result = create_prefix_lists(tgen, input_dict_2)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Create route map
        for addr_type in ADDR_TYPES:
            input_dict_3 = {
                "r3": {
                    "route_maps": {
                        "rmap_match_pf_1_{}".format(addr_type): [
                            {
                                "action": rmap_action,
                                "match": {
                                    addr_type: {
                                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                                    }
                                },
                            }
                        ]
                    }
                }
            }
            result = create_route_maps(tgen, input_dict_3)
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )

        # Configure neighbor for route map
        input_dict_7 = {
            "r3": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1_ipv4",
                                                        "direction": "in",
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1_ipv6",
                                                        "direction": "in",
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        },
                    }
                }
            }
        }

        result = create_router_bgp(tgen, topo, input_dict_7)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        dut = "r3"
        protocol = "bgp"
        input_dict_2 = {
            "r1": {
                "static_routes": [
                    {
                        "network": [NETWORK[adt][0]],
                        "no_of_ip": 9,
                        "next_hop": NEXT_HOP[adt],
                    }
                ]
            }
        }

        # tgen.mininet_cli()
        if "deny" in [prefix_action, rmap_action]:
            result = verify_rib(
                tgen, adt, dut, input_dict_2, protocol=protocol, expected=False
            )
            assert result is not True, (
                "Testcase {} : Failed \n "
                "Expected: Routes should not be present in {} FIB \n "
                "Found: {}".format(tc_name, dut, result)
            )
        else:
            result = verify_rib(tgen, adt, dut, input_dict_2, protocol=protocol)
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )


def test_route_map_multiple_seq_different_match_set_clause_p0(request):
    """
    TC_35:
    Test multiple sequence numbers in a single route-map for different
    match/set clauses.
    """

    tgen = get_topogen()
    # test case name
    tc_name = request.node.name
    write_test_header(tc_name)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    for adt in ADDR_TYPES:
        # Create Static routes
        input_dict = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[adt][0],
                        "no_of_ip": 1,
                        "next_hop": NEXT_HOP[adt],
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Api call to redistribute static routes
        input_dict_1 = {
            "r1": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Create ip prefix list
        input_dict_2 = {
            "r3": {
                "prefix_lists": {
                    "ipv4": {
                        "pf_list_1_ipv4": [
                            {"seqid": 10, "network": "any", "action": "permit"}
                        ]
                    },
                    "ipv6": {
                        "pf_list_1_ipv6": [
                            {"seqid": 100, "network": "any", "action": "permit"}
                        ]
                    },
                }
            }
        }
        result = create_prefix_lists(tgen, input_dict_2)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Create route map
        for addr_type in ADDR_TYPES:
            input_dict_3 = {
                "r3": {
                    "route_maps": {
                        "rmap_match_pf_1_{}".format(addr_type): [
                            {
                                "action": "permit",
                                "match": {
                                    addr_type: {
                                        "prefix_lists": "pf_list_2_{}".format(addr_type)
                                    }
                                },
                                "set": {"path": {"as_num": 500}},
                            },
                            {
                                "action": "permit",
                                "match": {
                                    addr_type: {
                                        "prefix_lists": "pf_list_2_{}".format(addr_type)
                                    }
                                },
                                "set": {
                                    "locPrf": 150,
                                },
                            },
                            {
                                "action": "permit",
                                "match": {
                                    addr_type: {
                                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                                    }
                                },
                                "set": {"metric": 50},
                            },
                        ]
                    }
                }
            }
            result = create_route_maps(tgen, input_dict_3)
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )

        # Configure neighbor for route map
        input_dict_4 = {
            "r3": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1_ipv4",
                                                        "direction": "in",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                    "r4": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1_ipv4",
                                                        "direction": "out",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                }
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1_ipv6",
                                                        "direction": "in",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                    "r4": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1_ipv6",
                                                        "direction": "out",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                }
                            }
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_4)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    for adt in ADDR_TYPES:
        # Verifying RIB routes
        dut = "r3"
        protocol = "bgp"
        input_dict = {
            "r3": {
                "route_maps": {
                    "rmap_match_pf_list1": [
                        {
                            "set": {
                                "metric": 50,
                            }
                        }
                    ],
                }
            }
        }

        static_routes = [NETWORK[adt][0]]

        time.sleep(2)
        result = verify_bgp_attributes(
            tgen, adt, dut, static_routes, "rmap_match_pf_list1", input_dict
        )
        assert result is True, "Test case {} : Failed \n Error: {}".format(
            tc_name, result
        )

        dut = "r4"
        result = verify_bgp_attributes(
            tgen, adt, dut, static_routes, "rmap_match_pf_list1", input_dict
        )
        assert result is True, "Test case {} : Failed \n Error: {}".format(
            tc_name, result
        )

        logger.info("Testcase " + tc_name + " :Passed \n")

        # Uncomment next line for debugging
        # tgen.mininet_cli()


def test_route_map_set_only_no_match_p0(request):
    """
    TC_37:
    Test add/remove route-maps with multiple set
    clauses and without any match statement.(Set only)
    """

    tgen = get_topogen()
    # test case name
    tc_name = request.node.name
    write_test_header(tc_name)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    for adt in ADDR_TYPES:
        # Create Static routes
        input_dict = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[adt][0],
                        "no_of_ip": 1,
                        "next_hop": NEXT_HOP[adt],
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Api call to redistribute static routes
        input_dict_1 = {
            "r1": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Create route map
        input_dict_3 = {
            "r3": {
                "route_maps": {
                    "rmap_match_pf_1": [
                        {
                            "action": "permit",
                            "set": {"metric": 50, "locPrf": 150, "weight": 4000},
                        }
                    ]
                }
            }
        }
        result = create_route_maps(tgen, input_dict_3)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Configure neighbor for route map
        input_dict_4 = {
            "r3": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1",
                                                        "direction": "in",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                    "r4": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1",
                                                        "direction": "out",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                }
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1",
                                                        "direction": "in",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                    "r4": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1",
                                                        "direction": "out",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                }
                            }
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_4)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    time.sleep(2)
    for adt in ADDR_TYPES:
        input_dict_4 = {
            "r3": {
                "route_maps": {
                    "rmap_match_pf_1": [
                        {
                            "action": "permit",
                            "set": {
                                "metric": 50,
                            },
                        }
                    ]
                }
            }
        }
        # Verifying RIB routes
        static_routes = [NETWORK[adt][0]]
        result = verify_bgp_attributes(
            tgen, adt, "r3", static_routes, "rmap_match_pf_1", input_dict_3
        )
        assert result is True, "Test case {} : Failed \n Error: {}".format(
            tc_name, result
        )

        result = verify_bgp_attributes(
            tgen, adt, "r4", static_routes, "rmap_match_pf_1", input_dict_4
        )
        assert result is True, "Test case {} : Failed \n Error: {}".format(
            tc_name, result
        )

        logger.info("Testcase " + tc_name + " :Passed \n")

        # Uncomment next line for debugging
        # tgen.mininet_cli()


def test_route_map_match_only_no_set_p0(request):
    """
    TC_38:
    Test add/remove route-maps with multiple match
    clauses and without any set statement.(Match only)
    """

    tgen = get_topogen()
    # test case name
    tc_name = request.node.name
    write_test_header(tc_name)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    for adt in ADDR_TYPES:
        # Create Static routes
        input_dict = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[adt][0],
                        "no_of_ip": 1,
                        "next_hop": NEXT_HOP[adt],
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Api call to redistribute static routes
        input_dict_1 = {
            "r1": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "redistribute": [
                                    {"redist_type": "static"},
                                    {"redist_type": "connected"},
                                ]
                            }
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Create ip prefix list
        input_dict_2 = {
            "r1": {
                "prefix_lists": {
                    "ipv4": {
                        "pf_list_1_ipv4": [
                            {"seqid": 10, "network": "any", "action": "permit"}
                        ]
                    },
                    "ipv6": {
                        "pf_list_1_ipv6": [
                            {"seqid": 100, "network": "any", "action": "permit"}
                        ]
                    },
                }
            }
        }
        result = create_prefix_lists(tgen, input_dict_2)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Create route map
        for addr_type in ADDR_TYPES:
            input_dict_3 = {
                "r1": {
                    "route_maps": {
                        "rmap_match_pf_1_{}".format(addr_type): [
                            {
                                "action": "permit",
                                "set": {
                                    "metric": 50,
                                    "locPrf": 150,
                                },
                            }
                        ]
                    }
                }
            }
            result = create_route_maps(tgen, input_dict_3)
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )

        # Configure neighbor for route map
        input_dict_4 = {
            "r1": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "neighbor": {
                                    "r3": {
                                        "dest_link": {
                                            "r1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1_ipv4",
                                                        "direction": "out",
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r3": {
                                        "dest_link": {
                                            "r1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_1_ipv6",
                                                        "direction": "out",
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_4)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Create ip prefix list
        input_dict_5 = {
            "r3": {
                "prefix_lists": {
                    "ipv4": {
                        "pf_list_1_ipv4": [
                            {"seqid": 10, "network": "any", "action": "permit"}
                        ]
                    },
                    "ipv6": {
                        "pf_list_1_ipv6": [
                            {"seqid": 100, "network": "any", "action": "permit"}
                        ]
                    },
                }
            }
        }
        result = create_prefix_lists(tgen, input_dict_5)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        # Create route map
        for addr_type in ADDR_TYPES:
            input_dict_6 = {
                "r3": {
                    "route_maps": {
                        "rmap_match_pf_2_{}".format(addr_type): [
                            {
                                "action": "permit",
                                "match": {
                                    addr_type: {
                                        "prefix_lists": "pf_list_1_{}".format(addr_type)
                                    }
                                },
                            }
                        ]
                    }
                }
            }
            result = create_route_maps(tgen, input_dict_6)
            assert result is True, "Testcase {} : Failed \n Error: {}".format(
                tc_name, result
            )

        # Configure neighbor for route map
        input_dict_7 = {
            "r3": {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_2_ipv4",
                                                        "direction": "in",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                    "r4": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_2_ipv4",
                                                        "direction": "out",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                }
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_2_ipv6",
                                                        "direction": "in",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                    "r4": {
                                        "dest_link": {
                                            "r3": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_match_pf_2_ipv6",
                                                        "direction": "out",
                                                    }
                                                ]
                                            }
                                        }
                                    },
                                }
                            }
                        },
                    }
                }
            }
        }
        result = create_router_bgp(tgen, topo, input_dict_7)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    for adt in ADDR_TYPES:
        # Verifying RIB routes
        static_routes = [NETWORK[adt][0]]
        result = verify_bgp_attributes(
            tgen, adt, "r3", static_routes, "rmap_match_pf_1", input_dict_3
        )
        assert result is True, "Test case {} : Failed \n Error: {}".format(
            tc_name, result
        )


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# SPDX-License-Identifier: ISC
#
# topolog.py
# Library of helper functions for NetDEF Topology Tests
#
# Copyright (c) 2017 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
Logging utilities for topology tests.

This file defines our logging abstraction.
"""

import logging
import os

try:
    from xdist import is_xdist_controller
except ImportError:

    def is_xdist_controller():
        return False


# Helper dictionary to convert Topogen logging levels to Python's logging.
DEBUG_TOPO2LOGGING = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "output": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
FORMAT = "%(asctime)s %(levelname)s: %(name)s: %(message)s"

handlers = {}
logger = logging.getLogger("topo")


# Remove this and use munet version when we move to pytest_asyncio
def get_test_logdir(nodeid=None, module=False):
    """Get log directory relative pathname."""
    xdist_worker = os.getenv("PYTEST_XDIST_WORKER", "")
    mode = os.getenv("PYTEST_XDIST_MODE", "no")

    # nodeid: all_protocol_startup/test_all_protocol_startup.py::test_router_running
    # may be missing "::testname" if module is True
    if not nodeid:
        nodeid = os.environ["PYTEST_CURRENT_TEST"].split(" ")[0]

    cur_test = nodeid.replace("[", "_").replace("]", "_")
    if module:
        idx = cur_test.rfind("::")
        path = cur_test if idx == -1 else cur_test[:idx]
        testname = ""
    else:
        path, testname = cur_test.split("::")
        testname = testname.replace("/", ".")
    path = path[:-3].replace("/", ".")

    # We use different logdir paths based on how xdist is running.
    if mode == "each":
        if module:
            return os.path.join(path, "worker-logs", xdist_worker)
        return os.path.join(path, testname, xdist_worker)
    assert mode in ("no", "load", "loadfile", "loadscope"), f"Unknown dist mode {mode}"
    return path if module else os.path.join(path, testname)


def set_handler(lg, target=None):
    if target is None:
        h = logging.NullHandler()
    else:
        if isinstance(target, str):
            h = logging.FileHandler(filename=target, mode="w")
        else:
            h = logging.StreamHandler(stream=target)
        h.setFormatter(logging.Formatter(fmt=FORMAT))
    # Don't filter anything at the handler level
    h.setLevel(logging.DEBUG)
    lg.addHandler(h)
    return h


def set_log_level(lg, level):
    "Set the logging level."
    # Messages sent to this logger only are created if this level or above.
    log_level = DEBUG_TOPO2LOGGING.get(level, level)
    lg.setLevel(log_level)


def reset_logger(lg):
    while lg.handlers:
        x = lg.handlers.pop()
        x.close()
        lg.removeHandler(x)


def get_logger(name, log_level=None, target=None, reset=True):
    lg = logging.getLogger(name)

    if reset:
        reset_logger(lg)

    if log_level is not None:
        set_log_level(lg, log_level)

    if target is not None:
        set_handler(lg, target)

    return lg


def logstart(nodeid, logpath):
    """Called from pytest before module setup."""
    worker = os.getenv("PYTEST_TOPOTEST_WORKER", "")
    wstr = f" on worker {worker}" if worker else ""
    handler_id = nodeid + worker
    logpath = logpath.absolute()

    logging.debug("logstart: adding logging for %s%s at %s", nodeid, wstr, logpath)
    root_logger = logging.getLogger()
    handler = logging.FileHandler(logpath, mode="w")
    handler.setFormatter(logging.Formatter(FORMAT))

    root_logger.addHandler(handler)
    handlers[handler_id] = handler

    logging.debug("logstart: added logging for %s%s at %s", nodeid, wstr, logpath)
    return handler


def logfinish(nodeid, logpath):
    """Called from pytest after module teardown."""
    worker = os.getenv("PYTEST_TOPOTEST_WORKER", "")
    wstr = f" on worker {worker}" if worker else ""

    root_logger = logging.getLogger()

    handler_id = nodeid + worker

    if handler_id not in handlers:
        logging.critical("can't find log handler to remove")
    else:
        logging.debug(
            "logfinish: removing logging for %s%s at %s", nodeid, wstr, logpath
        )
        h = handlers[handler_id]
        root_logger.removeHandler(h)
        h.flush()
        h.close()
        del handlers[handler_id]
        logging.debug(
            "logfinish: removed logging for %s%s at %s", nodeid, wstr, logpath
        )


console_handler = set_handler(logger, None)
set_log_level(logger, "debug")



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_high_ecmp.py
#
# Copyright (c) 2024 by
# Nvidia Corporation
# Donald Sharp
#
# Copyright (c) 2025 by Soumya Roy, <souroy@nvidia.com>

"""
test_high_ecmp.py: Testing two routers with 256 interfaces and BGP setup
                   on it.

"""

import os
import re
import sys
import pytest
import json

pytestmark = [pytest.mark.bgpd]

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.

#####################################################
##
##   Network Topology Definition
##
#####################################################


def build_topo(tgen):
    tgen.add_router("r1")
    tgen.add_router("r2")

    r1 = tgen.gears["r1"]
    r2 = tgen.gears["r2"]

    # Let's create 257 interfaces between the two switches
    for switch in range(1, 516):
        switch = tgen.add_switch("sw{}".format(switch))
        switch.add_link(r1)
        switch.add_link(r2)


#####################################################
##
##   Tests starting
##
#####################################################


def setup_module(module):
    "Setup topology"
    tgen = Topogen(build_topo, module.__name__)
    tgen.start_topology()

    # This is a sample of configuration loading.
    router_list = tgen.routers()
    for rname, router in router_list.items():
        router.load_frr_config(
            os.path.join(CWD, "{}/frr.conf".format(rname)),
            [
                (TopoRouter.RD_ZEBRA, "-s 180000000"),
                (TopoRouter.RD_BGP, None),
                (TopoRouter.RD_SHARP, None),
            ],
        )
 
    tgen.start_router()
    
    for rname, router in router_list.items():
       router.cmd("vtysh -f {}/{}/frr_ipv6_bgp.conf".format(CWD, rname))

def teardown_module(_mod):
    "Teardown the pytest environment"
    tgen = get_topogen()

    # This function tears down the whole topology.
    tgen.stop_topology()


def test_nothing():
    "Do Nothing"
    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_zebra_netlink.py
#
# Copyright (c) 2020 by
#

"""
test_zebra_netlink.py: Test some basic interactions with kernel using Netlink

"""
# pylint: disable=C0413
import ipaddress
import sys

import pytest
from lib import topotest
from lib.topogen import Topogen, TopoRouter
from lib.topolog import logger

pytestmark = [pytest.mark.sharpd]


#####################################################
##
##   Tests starting
##
#####################################################


@pytest.fixture(scope="module")
def tgen(request):
    "Sets up the pytest environment"

    topodef = {"s1": ("r1")}
    tgen = Topogen(topodef, request.module.__name__)
    tgen.start_topology()

    # Initialize all routers.
    router_list = tgen.routers()
    for _, router in router_list.items():
        router.load_config(TopoRouter.RD_ZEBRA, "zebra.conf")
        router.load_config(TopoRouter.RD_SHARP)

    tgen.start_router()
    yield tgen
    tgen.stop_topology()


@pytest.fixture(autouse=True)
def skip_on_failure(tgen):
    if tgen.routers_have_failure():
        pytest.skip("skipped because of previous test failure")


def test_zebra_netlink_batching(tgen):
    "Test the situation where dataplane fills netlink send buffer entirely."
    logger.info(
        "Test the situation where dataplane fills netlink send buffer entirely."
    )
    r1 = tgen.gears["r1"]

    # Reduce the size of the buffer to hit the limit.
    r1.vtysh_cmd("conf t\nzebra kernel netlink batch-tx-buf 256 256")

    entry = {"r1-eth0": {"addresses": ["192.168.1.1/24"]}}
    ok = topotest.router_json_cmp_retry(r1, "show int brief json", entry, False, 30)
    assert ok, '"r1" Address not installed yet'

    count = 100
    r1.vtysh_cmd("sharp install routes 2.1.3.7 nexthop 192.168.1.1 " + str(count))

    # Generate expected results
    entry = {
        "protocol": "sharp",
        "distance": 150,
        "metric": 0,
        "installed": True,
        "table": 254,
        "nexthops": [
            {
                "fib": True,
                "ip": "192.168.1.1",
                "afi": "ipv4",
                "interfaceName": "r1-eth0",
                "active": True,
                "weight": 1,
            }
        ],
    }

    match = {}
    base = int(ipaddress.ip_address("2.1.3.7"))
    for i in range(base, base + count):
        pfx = str(ipaddress.ip_network((i, 32)))
        match[pfx] = [dict(entry, prefix=pfx)]

    ok = topotest.router_json_cmp_retry(r1, "show ip route json", match, False, 30)
    assert ok, '"r1" JSON output mismatches'

    r1.vtysh_cmd("sharp remove routes 2.1.3.7 " + str(count))


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
